from typing import Dict

from PySide2 import QtWidgets, QtCore

from randovania.game_description.game_description import GameDescription
from randovania.games.game import RandovaniaGame
from randovania.gui.generated.preset_starting_area_ui import Ui_PresetStartingArea
from randovania.gui.lib.area_list_helper import AreaListHelper
from randovania.gui.preset_settings.preset_tab import PresetTab
from randovania.interface_common.preset_editor import PresetEditor
from randovania.layout.base.base_configuration import StartingLocationList
from randovania.layout.preset import Preset


class PresetStartingArea(PresetTab, Ui_PresetStartingArea, AreaListHelper):

    _starting_location_for_world: Dict[str, QtWidgets.QCheckBox]
    _starting_location_for_area: Dict[int, QtWidgets.QCheckBox]

    def __init__(self, editor: PresetEditor, game: GameDescription):
        super().__init__(editor)
        self.setupUi(self)
        self.game_description = game

        self.starting_area_layout.setAlignment(QtCore.Qt.AlignTop)

        self._starting_location_for_world, self._starting_location_for_area = self.create_area_list_selection(
            self.starting_locations_contents,
            self.starting_locations_layout,
            StartingLocationList.areas_list(self.game_description.game),
            self._on_starting_area_check_changed,
        )
        self.starting_area_quick_fill_ship.clicked.connect(self._starting_location_on_select_ship)
        self.starting_area_quick_fill_save_station.clicked.connect(self._starting_location_on_select_save_station)

    @property
    def uses_patches_tab(self) -> bool:
        return True

    @property
    def game_enum(self) -> RandovaniaGame:
        return self.game_description.game

    def _on_starting_area_check_changed(self, world_areas, checked: bool):
        with self._editor as editor:
            editor.set_configuration_field(
                "starting_location",
                editor.configuration.starting_location.ensure_has_locations(world_areas, checked)
            )

    def _starting_location_on_select_ship(self):
        with self._editor as editor:
            editor.set_configuration_field(
                "starting_location",
                editor.configuration.starting_location.with_elements(
                    [self.game_description.starting_location],
                    self.game_enum,
                )
            )

    def _starting_location_on_select_save_station(self):
        world_list = self.game_description.world_list
        save_stations = [world_list.node_to_area_location(node)
                         for node in world_list.all_nodes if node.name == "Save Station"]

        with self._editor as editor:
            editor.set_configuration_field(
                "starting_location",
                editor.configuration.starting_location.with_elements(save_stations, self.game_enum)
            )

    def on_preset_changed(self, preset: Preset):
        self.update_area_list(
            preset.configuration.starting_location.locations,
            False,
            self._starting_location_for_world,
            self._starting_location_for_area,
        )



from typing import List, Dict, Tuple, Optional

from randovania.game_description.game_patches import GamePatches
from randovania.game_description.resources import resource_info
from randovania.game_description.resources.item_resource_info import ItemResourceInfo
from randovania.game_description.resources.pickup_index import PickupIndex
from randovania.games.prime.patcher_file_lib import hint_lib
from randovania.interface_common.players_configuration import PlayersConfiguration


def find_locations_that_gives_items(
        target_items: List[ItemResourceInfo],
        all_patches: Dict[int, GamePatches],
        player: int,
) -> Dict[ItemResourceInfo, List[Tuple[int, PickupIndex]]]:
    result = {item: [] for item in target_items}

    for other_player, patches in all_patches.items():
        for pickup_index, target in patches.pickup_assignment.items():
            if target.player != player:
                continue

            # TODO: iterate over all tiers of progression
            resources = resource_info.convert_resource_gain_to_current_resources(target.pickup.resource_gain({}))

            for resource, quantity in resources.items():
                if quantity > 0 and resource in result:
                    result[resource].append((other_player, pickup_index))

    return result


def hint_text_if_items_are_starting(
        target_items: List[ItemResourceInfo],
        all_patches: Dict[int, GamePatches],
        player: int,
        resource_color: hint_lib.TextColor,
) -> Dict[ItemResourceInfo, str]:
    result = {}

    for resource, quantity in all_patches[player].starting_items.items():
        if quantity > 0 and resource in target_items:
            result[resource] = "{} has no need to be located.".format(
                hint_lib.color_text(resource_color, resource.long_name))

    return result


def create_guaranteed_hints_for_resources(
        all_patches: Dict[int, GamePatches],
        players_config: PlayersConfiguration,
        area_namers: Dict[int, hint_lib.AreaNamer],
        hide_area: bool,
        items: List[ItemResourceInfo],
        resource_color: hint_lib.TextColor,
        location_color: hint_lib.TextColor,
) -> Dict[ItemResourceInfo, str]:
    """
    Creates a hint for where each of the given resources for the given player can be found, across all players.
    If the player starts with the resource, indicate such. Errors if any resource can't be found.
    Intended for Sky Temple Key/Artifacts hints.

    :param all_patches:
    :param players_config:
    :param area_namers: Area namer for all players in the LayoutDescription
    :param hide_area: Should the area of the location be hidden?
    :param items: The item resources to hint
    :param resource_color: The color to use for resource names.
    :param location_color: The color to use for location names.
    :return:
    """
    resulting_hints = hint_text_if_items_are_starting(items, all_patches, players_config.player_index, resource_color)
    locations_for_items = find_locations_that_gives_items(items, all_patches, players_config.player_index)

    used_locations = set()
    for resource, locations in locations_for_items.items():
        if resource in resulting_hints:
            continue

        location: Optional[Tuple[int, PickupIndex]] = None
        for option in locations:
            if option not in used_locations:
                location = option
                break

        if location is not None:
            used_locations.add(location)

            if players_config.is_multiworld:
                determiner = hint_lib.player_determiner(players_config, location[0])
            else:
                determiner = hint_lib.Determiner("", False)

            resulting_hints[resource] = "{} is located in {}{}.".format(
                hint_lib.color_text(resource_color, resource.long_name),
                determiner,
                area_namers[location[0]].location_name(location[1], hide_area, color=location_color),
            )

    if len(resulting_hints) != len(items):
        raise ValueError(
            "Expected to find {} between pickup placement and starting items, found {}".format(
                len(items),
                len(resulting_hints)
            ))

    return resulting_hints



import pytest

from randovania.game_description.item.ammo import AMMO_ITEM_CATEGORY, Ammo
from randovania.game_description.item.item_category import USELESS_ITEM_CATEGORY, ItemCategory
from randovania.game_description.item.major_item import MajorItem
from randovania.game_description.resources.pickup_entry import (
    PickupEntry,
    ResourceLock,
    PickupModel,
)
from randovania.game_description.resources.resource_info import add_resource_gain_to_current_resources
from randovania.generator.item_pool import pickup_creator
from randovania.layout.base.major_item_state import MajorItemState


@pytest.mark.parametrize("percentage", [False, True])
def test_create_pickup_for(percentage: bool, echoes_item_database, echoes_resource_database, generic_item_category):
    # Setup
    item_a = echoes_resource_database.get_item(10)
    item_b = echoes_resource_database.get_item(15)
    item_c = echoes_resource_database.get_item(18)
    ammo_a = echoes_resource_database.get_item(42)
    ammo_b = echoes_resource_database.get_item(45)

    less_generic_item_category = ItemCategory(
        name="the_category",
        long_name="The Category",
        hint_details=("a ", " wonderful item"),
        is_major=True
    )

    major_item = MajorItem(
        game=echoes_resource_database.game_enum,
        name="The Item",
        item_category=less_generic_item_category,
        broad_category=generic_item_category,
        model_name="SuperModel",
        progression=(10, 15, 18),
        ammo_index=(42, 45),
        required=False,
        original_index=None,
        probability_offset=5,
    )
    state = MajorItemState(
        include_copy_in_original_location=False,
        num_shuffled_pickups=0,
        num_included_in_starting_items=0,
        included_ammo=(10, 20),
    )

    if percentage:
        extra_resources = (
            (ammo_a, 10),
            (ammo_b, 20),
            (echoes_resource_database.item_percentage, 1),
        )
    else:
        extra_resources = (
            (ammo_a, 10),
            (ammo_b, 20),
        )

    # Run
    result = pickup_creator.create_major_item(major_item, state, percentage,
                                              echoes_resource_database,
                                              None, False)

    # Assert
    assert result == PickupEntry(
        name="The Item",
        model=PickupModel(echoes_resource_database.game_enum, "SuperModel"),
        progression=(
            (item_a, 1),
            (item_b, 1),
            (item_c, 1),
        ),
        extra_resources=extra_resources,
        item_category=less_generic_item_category,
        broad_category=generic_item_category,
        probability_offset=5,
        respects_lock=False,
    )


@pytest.mark.parametrize(["ammo_quantity"], [
    (0,),
    (10,),
    (15,),
])
def test_create_missile_launcher(ammo_quantity: int, echoes_item_database, echoes_resource_database):
    # Setup
    missile = echoes_resource_database.get_item(44)
    missile_launcher = echoes_resource_database.get_item(73)
    temporary = echoes_resource_database.get_item(71)

    state = MajorItemState(
        include_copy_in_original_location=False,
        num_shuffled_pickups=0,
        num_included_in_starting_items=0,
        included_ammo=(ammo_quantity,),
    )

    # Run
    result = pickup_creator.create_major_item(
        echoes_item_database.major_items["Missile Launcher"],
        state,
        True,
        echoes_resource_database,
        echoes_item_database.ammo["Missile Expansion"],
        True
    )

    # Assert
    assert result == PickupEntry(
        name="Missile Launcher",
        progression=(
            (missile_launcher, 1),
        ),
        extra_resources=(
            (missile, ammo_quantity),
            (echoes_resource_database.item_percentage, 1),
        ),
        model=PickupModel(echoes_resource_database.game_enum, "MissileLauncher"),
        item_category=echoes_item_database.item_categories["missile"],
        broad_category=echoes_item_database.item_categories["missile_related"],
        resource_lock=ResourceLock(
            locked_by=missile_launcher,
            temporary_item=temporary,
            item_to_lock=missile,
        ),
        unlocks_resource=True,
    )


@pytest.mark.parametrize("ammo_quantity", [0, 10, 15])
@pytest.mark.parametrize("ammo_requires_major_item", [False, True])
def test_create_seeker_launcher(ammo_quantity: int,
                                ammo_requires_major_item: bool,
                                echoes_item_database,
                                echoes_resource_database,
                                ):
    # Setup
    missile = echoes_resource_database.get_item(44)
    missile_launcher = echoes_resource_database.get_item(73)
    seeker_launcher = echoes_resource_database.get_item(26)
    temporary = echoes_resource_database.get_item(71)

    state = MajorItemState(
        include_copy_in_original_location=False,
        num_shuffled_pickups=0,
        num_included_in_starting_items=0,
        included_ammo=(ammo_quantity,),
    )

    # Run
    result = pickup_creator.create_major_item(
        echoes_item_database.major_items["Seeker Launcher"],
        state,
        True,
        echoes_resource_database,
        echoes_item_database.ammo["Missile Expansion"],
        ammo_requires_major_item
    )

    # Assert

    assert result == PickupEntry(
        name="Seeker Launcher",
        progression=(
            (seeker_launcher, 1),
        ),
        extra_resources=(
            (missile, ammo_quantity),
            (echoes_resource_database.item_percentage, 1),
        ),
        model=PickupModel(echoes_resource_database.game_enum, "SeekerLauncher"),
        item_category=echoes_item_database.item_categories["missile"],
        broad_category=echoes_item_database.item_categories["missile_related"],
        respects_lock=ammo_requires_major_item,
        resource_lock=ResourceLock(
            locked_by=missile_launcher,
            temporary_item=temporary,
            item_to_lock=missile,
        ),
    )


@pytest.mark.parametrize("requires_major_item", [False, True])
def test_create_ammo_expansion(requires_major_item: bool, echoes_item_database, echoes_resource_database):
    # Setup
    primary_a = echoes_resource_database.get_item(73)
    ammo_a = echoes_resource_database.get_item(42)
    temporary_a = echoes_resource_database.get_item(71)

    ammo = Ammo(
        game=echoes_resource_database.game_enum,
        name="The Item",
        items=(42,),
        broad_category=USELESS_ITEM_CATEGORY,
        unlocked_by=73,
        temporary=71,
        model_name="AmmoModel",
    )
    ammo_count = (11, 150)

    # Run
    result = pickup_creator.create_ammo_expansion(
        ammo, ammo_count, requires_major_item, echoes_resource_database)

    # Assert
    assert result == PickupEntry(
        name="The Item",
        model=PickupModel(echoes_resource_database.game_enum, "AmmoModel"),
        progression=tuple(),
        extra_resources=(
            (ammo_a, ammo_count[0]),
            (echoes_resource_database.item_percentage, 1),
        ),
        item_category=AMMO_ITEM_CATEGORY,
        broad_category=USELESS_ITEM_CATEGORY,
        probability_offset=0,
        respects_lock=requires_major_item,
        resource_lock=ResourceLock(
            locked_by=primary_a,
            temporary_item=temporary_a,
            item_to_lock=ammo_a,
        ),
    )


@pytest.mark.parametrize("include_before", [False, True])
def test_missile_expansion_before_launcher(include_before, echoes_item_database, echoes_resource_database):
    # Setup
    ammo = echoes_item_database.ammo["Missile Expansion"]
    major_item = echoes_item_database.major_items["Missile Launcher"]

    missile = echoes_resource_database.get_item(ammo.items[0])
    missile_launcher = echoes_resource_database.get_item(major_item.progression[0])
    temporary = echoes_resource_database.get_item(ammo.temporary)
    percent = echoes_resource_database.item_percentage

    # Run
    expansion = pickup_creator.create_ammo_expansion(ammo, [5], True, echoes_resource_database)
    launcher = pickup_creator.create_major_item(
        major_item, MajorItemState(included_ammo=(5,)),
        True, echoes_resource_database, ammo, True
    )

    resources = {}

    if include_before:
        # Ammo Expansion
        add_resource_gain_to_current_resources(expansion.resource_gain(resources, force_lock=True), resources)
        assert resources == {percent: 1, temporary: 5}

    # Add Launcher
    add_resource_gain_to_current_resources(launcher.resource_gain(resources, force_lock=True), resources)
    if include_before:
        assert resources == {percent: 2, temporary: 0, missile_launcher: 1, missile: 10}
    else:
        assert resources == {percent: 1, temporary: 0, missile_launcher: 1, missile: 5}

    # Ammo Expansion
    add_resource_gain_to_current_resources(expansion.resource_gain(resources, force_lock=True), resources)
    if include_before:
        assert resources == {percent: 3, temporary: 0, missile_launcher: 1, missile: 15}
    else:
        assert resources == {percent: 2, temporary: 0, missile_launcher: 1, missile: 10}



import dataclasses
import inspect
import typing
from enum import Enum


def _handle_optional(type_):
    if getattr(type_, "__origin__", None) is typing.Union and (
            len(type_.__args__) == 2 and type_.__args__[1] is type(None)):
        return type_.__args__[0]
    return type_


T = typing.TypeVar("T")


class JsonDataclass:
    @property
    def as_json(self) -> dict:
        result = {}
        for field in dataclasses.fields(self):
            if not field.init:
                continue
            value = getattr(self, field.name)
            if isinstance(value, Enum):
                value = value.value
            elif value is not None and hasattr(value, "as_json"):
                value = value.as_json
            result[field.name] = value
        return result

    @classmethod
    def json_extra_arguments(cls):
        return {}

    @classmethod
    def from_json(cls: typing.Type[T], json_dict: dict, **extra) -> T:
        extra_args = cls.json_extra_arguments()
        extra_args.update(extra)

        new_instance = {}
        for field in dataclasses.fields(cls):
            if not field.init or (field.name not in json_dict and (field.default != dataclasses.MISSING
                                                                   or field.default_factory != dataclasses.MISSING)):
                continue
            arg = json_dict[field.name]
            if arg is not None:
                type_ = _handle_optional(field.type)
                if issubclass(type_, Enum):
                    arg = type_(arg)
                elif hasattr(type_, "from_json"):
                    arg_spec = inspect.getfullargspec(type_.from_json)
                    arg = type_.from_json(arg, **{
                        name: value
                        for name, value in extra_args.items()
                        if arg_spec.varkw is not None or name in arg_spec.args or name in arg_spec.kwonlyargs
                    })
            new_instance[field.name] = arg

        unknown_keys = set(json_dict.keys()) - set(new_instance.keys())
        if unknown_keys:
            raise ValueError(f"Unknown keys present in argument: {unknown_keys}")

        return cls(**new_instance)



from dataclasses import dataclass
from typing import Dict, Tuple, Optional

from randovania.game_description.item.item_category import ItemCategory
from randovania.game_description.resources.pickup_entry import ResourceLock
from randovania.game_description.resources.resource_database import ResourceDatabase
from randovania.games.game import RandovaniaGame


@dataclass(frozen=True, order=True)
class Ammo:
    game: RandovaniaGame
    name: str
    model_name: str
    items: Tuple[int, ...]
    broad_category: ItemCategory
    unlocked_by: Optional[int] = None
    temporary: Optional[int] = None

    def __post_init__(self):
        if self.temporary is not None:
            if self.unlocked_by is None:
                raise ValueError("If temporaries is set, unlocked_by must be set.")
            if len(self.items) != 1:
                raise ValueError("If temporaries is set, only one item is supported. Got {0} instead".format(
                    len(self.items)
                ))
        elif self.unlocked_by is not None:
            raise ValueError("If temporaries is not set, unlocked_by must not be set.")

    @classmethod
    def from_json(cls, name: str, value: dict, game: RandovaniaGame,
                  item_categories: Dict[str, ItemCategory]) -> "Ammo":
        return cls(
            game=game,
            name=name,
            model_name=value["model_name"],
            items=tuple(value["items"]),
            broad_category=item_categories[value["broad_category"]],
            unlocked_by=value.get("unlocked_by"),
            temporary=value.get("temporary"),
        )

    @property
    def as_json(self) -> dict:
        result = {
            "model_name": self.model_name,
            "items": list(self.items),
            "broad_category": self.broad_category.name,
        }
        if self.unlocked_by is not None:
            result["temporary"] = self.temporary
            result["unlocked_by"] = self.unlocked_by
        return result

    @property
    def item_category(self) -> ItemCategory:
        return AMMO_ITEM_CATEGORY

    def create_resource_lock(self, resource_database: ResourceDatabase) -> Optional[ResourceLock]:
        if self.unlocked_by is not None:
            return ResourceLock(
                locked_by=resource_database.get_item(self.unlocked_by),
                item_to_lock=resource_database.get_item(self.items[0]),
                temporary_item=resource_database.get_item(self.temporary),
            )
        return None


AMMO_ITEM_CATEGORY = ItemCategory(
    name="expansion",
    long_name="",
    hint_details=("an ", "expansion"),
    is_major=False
)



import typing
from typing import Tuple, Callable, Dict

from randovania.game_description import default_database
from randovania.game_description.resources.resource_database import ResourceDatabase
from randovania.game_description.resources.resource_info import add_resources_into_another
from randovania.games.game import RandovaniaGame
from randovania.generator.item_pool import PoolResults
from randovania.generator.item_pool.ammo import add_ammo
from randovania.generator.item_pool.artifacts import add_artifacts
from randovania.generator.item_pool.dark_temple_keys import add_dark_temple_keys
from randovania.generator.item_pool.energy_cells import add_energy_cells
from randovania.generator.item_pool.major_items import add_major_items
from randovania.generator.item_pool.sky_temple_keys import add_sky_temple_key_distribution_logic
from randovania.layout.base.base_configuration import BaseConfiguration
from randovania.layout.prime3.corruption_configuration import CorruptionConfiguration
from randovania.layout.prime2.echoes_configuration import EchoesConfiguration
from randovania.layout.prime1.prime_configuration import PrimeConfiguration


def _extend_pool_results(base_results: PoolResults, extension: PoolResults):
    base_results.pickups.extend(extension.pickups)
    base_results.assignment.update(extension.assignment)
    add_resources_into_another(base_results.initial_resources, extension.initial_resources)


def prime1_specific_pool(results: PoolResults, configuration: PrimeConfiguration, db: ResourceDatabase):
    _extend_pool_results(results, add_artifacts(db, configuration.artifact_target,
                                                configuration.artifact_minimum_progression))


def echoes_specific_pool(results: PoolResults, configuration: EchoesConfiguration, db: ResourceDatabase):
    # Adding Dark Temple Keys to pool
    _extend_pool_results(results, add_dark_temple_keys(db))

    # Adding Sky Temple Keys to pool
    _extend_pool_results(results, add_sky_temple_key_distribution_logic(db, configuration.sky_temple_keys))


def corruption_specific_pool(results: PoolResults, configuration: CorruptionConfiguration, db: ResourceDatabase):
    # Adding Energy Cells to pool
    _extend_pool_results(results, add_energy_cells(db))


def super_metroid_specific_pool(results: PoolResults, configuration: CorruptionConfiguration, db: ResourceDatabase):
    pass


_GAME_SPECIFIC = typing.cast(Dict[RandovaniaGame, Callable[[PoolResults, BaseConfiguration, ResourceDatabase], None]], {
    RandovaniaGame.METROID_PRIME: prime1_specific_pool,
    RandovaniaGame.METROID_PRIME_ECHOES: echoes_specific_pool,
    RandovaniaGame.METROID_PRIME_CORRUPTION: corruption_specific_pool,
    RandovaniaGame.SUPER_METROID: super_metroid_specific_pool,
})


def calculate_pool_results(layout_configuration: BaseConfiguration,
                           resource_database: ResourceDatabase,
                           ) -> PoolResults:
    """
    Creates a PoolResults with all starting items and pickups in fixed locations, as well as a list of
    pickups we should shuffle.
    :param layout_configuration:
    :param resource_database:
    :return:
    """
    base_results = PoolResults([], {}, {})

    # Adding major items to the pool
    _extend_pool_results(base_results, add_major_items(resource_database,
                                                       layout_configuration.major_items_configuration,
                                                       layout_configuration.ammo_configuration))

    # Adding ammo to the pool
    base_results.pickups.extend(add_ammo(resource_database,
                                         layout_configuration.ammo_configuration))

    _GAME_SPECIFIC[layout_configuration.game](base_results, layout_configuration, resource_database)

    return base_results


def calculate_pool_item_count(layout: BaseConfiguration) -> Tuple[int, int]:
    """
    Calculate how many pickups are needed for given layout, with how many spots are there.
    :param layout:
    :return:
    """
    game_description = default_database.game_description_for(layout.game)
    num_pickup_nodes = game_description.world_list.num_pickup_nodes

    pool_pickups, pool_assignment, _ = calculate_pool_results(layout, game_description.resource_database)
    min_starting_items = layout.major_items_configuration.minimum_random_starting_items

    pool_count = len(pool_pickups) + len(pool_assignment)
    maximum_size = num_pickup_nodes + min_starting_items

    return pool_count, maximum_size



import argparse
import json
import typing
from argparse import ArgumentParser
from pathlib import Path
from typing import Dict, BinaryIO, Optional, TextIO, List, Any

from randovania.game_description.resources.resource_info import ResourceInfo
from randovania.game_description.resources.search import MissingResource, find_resource_info_with_long_name
from randovania.games import default_data, binary_data
from randovania.games.game import RandovaniaGame
from randovania.lib.enum_lib import iterate_enum


def _get_sorted_list_of_names(input_list: List[Any], prefix: str = "") -> List[str]:
    for item in sorted(input_list, key=lambda x: x.name):
        yield prefix + item.name


def decode_data_file(args) -> Dict:
    json_database: Optional[Path] = args.json_database
    if json_database is not None:
        with json_database.open() as data_file:
            return json.load(data_file)
    else:
        return default_data.read_json_then_binary(RandovaniaGame(args.game))[1]


def export_as_binary(data: dict, output_binary: Path):
    with output_binary.open("wb") as x:  # type: BinaryIO
        binary_data.encode(data, x)


def convert_database_command_logic(args):
    from randovania.game_description import data_reader, data_writer
    data = decode_data_file(args)

    if args.decode_to_game_description:
        data = data_writer.write_game_description(data_reader.decode_data(data))

    output_binary: Optional[Path] = args.output_binary
    output_json: Optional[Path] = args.output_json

    if output_binary is not None:
        export_as_binary(data, output_binary)

    elif output_json is not None:
        with output_json.open("w") as x:  # type: TextIO
            json.dump(data, x, indent=4)
    else:
        raise ValueError("Neither binary nor JSON set. Argparse is broken?")


def create_convert_database_command(sub_parsers):
    parser: ArgumentParser = sub_parsers.add_parser(
        "convert-database",
        help="Converts a database file between JSON and binary encoded formats. Input defaults to embedded database.",
        formatter_class=argparse.MetavarTypeHelpFormatter
    )
    parser.add_argument(
        "--decode-to-game-description",
        action="store_true",
        default=False,
        help="Decodes the input data to a GameDescription, then encodes it back."
    )

    group = parser.add_mutually_exclusive_group(required=True)
    group.add_argument(
        "--output-binary",
        type=Path,
        help="Export as a binary file.",
    )
    group.add_argument(
        "--output-json",
        type=Path,
        help="Export as a JSON file.",
    )

    parser.set_defaults(func=convert_database_command_logic)


def view_area_command_logic(args):
    from randovania.game_description import pretty_print
    game = load_game_description(args)
    world_list = game.world_list

    try:
        world = world_list.world_with_name(args.world)

    except KeyError:
        options = "\n".join(_get_sorted_list_of_names(world_list.worlds, " "))
        print(f"Unknown world named '{args.world}'. Options:\n{options}")
        raise SystemExit(1)

    try:
        area = world.area_by_name(args.area)

    except KeyError:
        options = "\n".join(_get_sorted_list_of_names(world.areas, " "))
        print(f"Unknown area named '{args.area}' in world {world}. Options:\n{options}")
        raise SystemExit(1)

    pretty_print.pretty_print_area(game, area)


def load_game_description(args):
    from randovania.game_description import data_reader

    data = decode_data_file(args)
    gd = data_reader.decode_data(data)
    return gd


def view_area_command(sub_parsers):
    parser: ArgumentParser = sub_parsers.add_parser(
        "view-area",
        help="View information about an area.",
        formatter_class=argparse.MetavarTypeHelpFormatter
    )
    parser.add_argument(
        "--simplify",
        action="store_true",
        help="Simplify the RequirementSets"
    )
    parser.add_argument(
        "world",
        type=str,
        help="The name of the world that contains the area."
    )
    parser.add_argument(
        "area",
        type=str,
        help="The name of the area."
    )

    parser.set_defaults(func=view_area_command_logic)


def update_human_readable_logic(args):
    from randovania.game_description import pretty_print
    from randovania.game_description import data_reader
    game = RandovaniaGame(args.game)

    path, data = default_data.read_json_then_binary(game)
    gd = data_reader.decode_data(data)

    path.with_suffix("").mkdir(parents=True, exist_ok=True)
    pretty_print.write_human_readable_game(gd, path.with_suffix(""))


def update_human_readable(sub_parsers):
    parser: ArgumentParser = sub_parsers.add_parser(
        "update-human-readable",
        help="Update the human readable versions",
        formatter_class=argparse.MetavarTypeHelpFormatter
    )
    parser.set_defaults(func=update_human_readable_logic)


def refresh_all_logic(args):
    from randovania.game_description import pretty_print
    from randovania.game_description import data_reader, data_writer

    for game in RandovaniaGame:
        path, data = default_data.read_json_then_binary(game)
        gd = data_reader.decode_data(data)
        new_data = data_writer.write_game_description(gd)

        data_writer.write_as_split_files(new_data, path)
        path.with_suffix("").mkdir(parents=True, exist_ok=True)
        pretty_print.write_human_readable_game(gd, path.with_suffix(""))


def refresh_all(sub_parsers):
    parser: ArgumentParser = sub_parsers.add_parser(
        "refresh-all",
        help="Re-exports the json and txt files of all databases",
        formatter_class=argparse.MetavarTypeHelpFormatter
    )
    parser.set_defaults(func=refresh_all_logic)


def _list_paths_with_resource(game,
                              print_only_area: bool,
                              resource: ResourceInfo,
                              needed_quantity: Optional[int]):
    from randovania.game_description.game_description import GameDescription

    count = 0
    game = typing.cast(GameDescription, game)

    for area in game.world_list.all_areas:
        area_had_resource = False

        for source, connection in area.connections.items():
            for target, requirement in connection.items():
                for alternative in requirement.as_set(game.resource_database).alternatives:
                    individual = alternative.get(resource)
                    if individual is None:
                        continue

                    if needed_quantity is None or needed_quantity == individual.amount:
                        area_had_resource = True
                        if not print_only_area:
                            print("At {0}, from {1} to {2}:\n{3}\n".format(
                                game.world_list.area_name(area),
                                source.name,
                                target.name,
                                sorted(individual for individual in alternative.values()
                                       if individual.resource != resource)
                            ))
                        count += 1

        if area_had_resource and print_only_area:
            print(game.world_list.area_name(area))

    print("Total routes: {}".format(count))


def list_paths_with_dangerous_logic(args):
    game = load_game_description(args)
    print_only_area = args.print_only_area
    count = 0

    for area in game.world_list.all_areas:
        area_had_resource = False

        for source, connection in area.connections.items():
            for target, requirement in connection.items():
                for alternative in requirement.as_set(game.resource_database).alternatives:
                    for individual in alternative.values():
                        if individual.negate:
                            area_had_resource = True
                            if not print_only_area:
                                print("At {0}, from {1} to {2}:\n{4}\n".format(
                                    game.world_list.area_name(area),
                                    area,
                                    source.name,
                                    target.name,
                                    sorted(individual for individual in alternative.values())))
                            count += 1

        if area_had_resource and print_only_area:
            print(game.world_list.area_name(area))

    print("Total routes: {}".format(count))


def list_paths_with_dangerous_command(sub_parsers):
    parser: ArgumentParser = sub_parsers.add_parser(
        "list-dangerous-usage",
        help="List all connections that needs a resource to be missing.",
        formatter_class=argparse.MetavarTypeHelpFormatter
    )
    parser.add_argument("--print-only-area", help="Only print the area names, not each specific path",
                        action="store_true")
    parser.set_defaults(func=list_paths_with_dangerous_logic)


def list_paths_with_resource_logic(args):
    gd = load_game_description(args)

    resource = None
    for resource_type in gd.resource_database:
        try:
            resource = find_resource_info_with_long_name(resource_type, args.resource)
            break
        except MissingResource:
            continue

    if resource is None:
        print(f"A resource named {args.resource} was not found.")
        raise SystemExit(1)

    _list_paths_with_resource(
        gd,
        args.print_only_area,
        resource,
        None
    )


def list_paths_with_resource_command(sub_parsers):
    parser: ArgumentParser = sub_parsers.add_parser(
        "list-resource-usage",
        help="List all connections that needs the resource.",
        formatter_class=argparse.MetavarTypeHelpFormatter
    )
    parser.add_argument("--print-only-area", help="Only print the area names, not each specific path",
                        action="store_true")
    parser.add_argument("resource", type=str)
    parser.set_defaults(func=list_paths_with_resource_logic)


def render_worlds_graph_logic(args):
    import hashlib
    import os
    import re
    import graphviz
    from randovania.game_description.world.node import PickupNode, DockNode, TeleporterNode

    gd = load_game_description(args)
    dot = graphviz.Digraph(comment=gd.game.long_name)

    worlds = list(gd.world_list.worlds)

    added_edges = set()
    _IMPOSSIBLE_LOCKS = {"No Return Portal", "Permanently Locked"}
    vulnerabilities_colors = {
        "Normal Door": None,
        "Morph Ball Door": None,
        "Other Door": None,
        "Scan Portal": None,

        "Missile": "#ff1919",
        "Super Missile": "#38c914",
        "Seeker Launcher": "#b233e8",
        "Power Bomb": "#dfe833",

        "Wave Door": "#a30af5",
        "Ice Door": "#7cdede",
        "Plasma Door": "#870f0f",

        "Light Door": "#bfd9d9",
        "Dark Door": "#3b3647",
        "Annihilator Door": "#616969",
        "Light Portal": "#bfd9d9",
        "Dark Portal": "#3b3647",
    }

    def _weakness_name(s: str):
        return re.sub(r"\([^)]*\)", "", s).replace(" Blast Shield", "").strip()

    def _hash_to_color(s: str) -> str:
        h = hashlib.blake2b(s.encode("utf-8"), digest_size=3).digest()
        return "#{:06x}".format(int.from_bytes(h, "big"))

    def _add_connection(dock_node: DockNode):
        the_world = gd.world_list.nodes_to_world(dock_node)
        source_area = gd.world_list.nodes_to_area(dock_node)
        target_node = gd.world_list.resolve_dock_connection(world, dock_node.default_connection)
        target_area = gd.world_list.nodes_to_area(target_node)

        if dock_node.default_dock_weakness.name in _IMPOSSIBLE_LOCKS:
            return

        edge_id = f"{the_world.world_asset_id}-{source_area.area_asset_id}-{dock_node.dock_index}"
        if edge_id in added_edges:
            return

        weak_name = _weakness_name(dock_node.default_dock_weakness.name)
        direction = None
        if isinstance(target_node, DockNode) and _weakness_name(target_node.default_dock_weakness.name) == weak_name:
            direction = "both"
            added_edges.add(f"{the_world.world_asset_id}-{target_area.area_asset_id}-{target_node.dock_index}")

        color = vulnerabilities_colors.get(weak_name, _hash_to_color(weak_name))
        dot.edge(
            f"{the_world.world_asset_id}-{source_area.area_asset_id}",
            f"{the_world.world_asset_id}-{target_area.area_asset_id}",
            weak_name, dir=direction, color=color, fontcolor=color,
        )
        added_edges.add(edge_id)

    def _add_teleporter(teleporter_node: TeleporterNode):
        source_world = gd.world_list.nodes_to_world(teleporter_node)
        source_area = gd.world_list.nodes_to_area(teleporter_node)
        target_node = gd.world_list.resolve_teleporter_connection(teleporter_node.default_connection)
        target_world = gd.world_list.nodes_to_world(target_node)
        target_area = gd.world_list.nodes_to_area(target_node)

        dot.edge(
            f"{source_world.world_asset_id}-{source_area.area_asset_id}",
            f"{target_world.world_asset_id}-{target_area.area_asset_id}",
            "Elevator",
        )

    per_game_colors = {
        RandovaniaGame.METROID_PRIME_ECHOES: {
            "Agon Wastes": "#ffc61c",
            "Torvus Bog": "#20ff1c",
            "Sanctuary Fortress": "#3d62ff",
            "Temple Grounds": "#c917ff",
            "Great Temple": "#c917ff",
        },
    }
    colors = per_game_colors.get(gd.game)
    if colors is None:
        colors = {
            world.name: _hash_to_color(world.name)
            for world in gd.world_list.worlds
        }

    dark_colors = {
        "Agon Wastes": "#a88332",
        "Torvus Bog": "#149612",
        "Sanctuary Fortress": "#112991",
        "Temple Grounds": "#7d2996",
        "Great Temple": "#7d2996",
    }

    for world in worlds:
        for area in world.areas:
            shape = None
            if any(isinstance(node, TeleporterNode) for node in area.nodes):
                shape = "polygon"

            c = (dark_colors if area.in_dark_aether else colors)[world.name]
            fillcolor = "".join("{:02x}".format(max(0, int(c[i * 2 + 1:i * 2 + 3], 16) - 64))
                                for i in range(3))
            dot.node(
                f"{world.world_asset_id}-{area.area_asset_id}", area.name,
                color=c,
                fillcolor=f"#{fillcolor}",
                style="filled",
                fontcolor=f"#ffffff",
                shape=shape,
                penwidth="3.0",
            )

            for node in area.nodes:
                if args.include_pickups and isinstance(node, PickupNode):
                    dot.node(str(node.pickup_index), re.search(r"Pickup \(([^)]+)\)", node.name).group(1),
                             shape="house")
                    dot.edge(f"{world.world_asset_id}-{area.area_asset_id}", str(node.pickup_index))

    for world in worlds:
        print(f"Adding docks for {world.name}")
        for area in world.areas:
            for node in area.nodes:
                if isinstance(node, DockNode):
                    _add_connection(node)
                elif isinstance(node, TeleporterNode) and node.editable and args.include_teleporters:
                    _add_teleporter(node)

    # os.environ["PATH"] += rf';C:\Program Files\Graphviz\bin'
    print(dot.render(format="png", cleanup=True, view=True))


def render_worlds_graph(sub_parsers):
    parser: ArgumentParser = sub_parsers.add_parser(
        "render-world-graph",
        help="Renders an image with all area connections",
        formatter_class=argparse.MetavarTypeHelpFormatter
    )
    parser.add_argument("--include-teleporters", action="store_true")
    parser.add_argument("--include-pickups", action="store_true")

    parser.set_defaults(func=render_worlds_graph_logic)


def pickups_per_area_command_logic(args):
    from randovania.game_description.world.node import PickupNode
    gd = load_game_description(args)

    for world in gd.world_list.worlds:
        num_pickups = sum(1 for node in world.all_nodes if isinstance(node, PickupNode))
        print(f"{world.correct_name(False)}: {num_pickups}")


def pickups_per_area_command(sub_parsers):
    parser: ArgumentParser = sub_parsers.add_parser(
        "pickups-per-area",
        help="Print how many pickups there are in each area",
        formatter_class=argparse.MetavarTypeHelpFormatter
    )
    parser.set_defaults(func=pickups_per_area_command_logic)


def create_subparsers(sub_parsers):
    parser: ArgumentParser = sub_parsers.add_parser(
        "database",
        help="Actions for database manipulation"
    )

    group = parser.add_mutually_exclusive_group()
    group.add_argument(
        "--game",
        type=str,
        choices=[game.value for game in iterate_enum(RandovaniaGame)],
        default=RandovaniaGame.METROID_PRIME_ECHOES.value,
        help="Use the included database for the given game.",
    )
    group.add_argument(
        "--json-database",
        type=Path,
        help="Path to the JSON encoded database.",
    )

    sub_parsers = parser.add_subparsers(dest="database_command")
    create_convert_database_command(sub_parsers)
    view_area_command(sub_parsers)
    update_human_readable(sub_parsers)
    refresh_all(sub_parsers)
    list_paths_with_dangerous_command(sub_parsers)
    list_paths_with_resource_command(sub_parsers)
    render_worlds_graph(sub_parsers)
    pickups_per_area_command(sub_parsers)

    def check_command(args):
        if args.database_command is None:
            parser.print_help()
            raise SystemExit(1)

    parser.set_defaults(func=check_command)



from test.conftest import generic_item_category
from unittest.mock import Mock

import pytest

from randovania.game_description.resources.item_resource_info import ItemResourceInfo
from randovania.game_description.resources.pickup_entry import PickupEntry, ResourceLock
from randovania.game_description.resources.pickup_index import PickupIndex
from randovania.resolver import state
from randovania.resolver.state import StateGameData


@pytest.fixture(name="state_game_data")
def _state_game_data() -> StateGameData:
    return StateGameData(
        Mock(energy_tank=ItemResourceInfo(42, "Energy Tank", "EnergyTank", 14, None),
             item_percentage=ItemResourceInfo(47, "Item Percentage", "Percentage", 255, None)),
        None,
        100
    )


def test_collected_pickup_indices(state_game_data, empty_patches):
    # Setup
    resources = {
        ItemResourceInfo(1, "A", "A", 50, None): 5,
        PickupIndex(1): 1,
        PickupIndex(15): 1
    }
    s = state.State(resources, (), 99, None, empty_patches, None, state_game_data)

    # Run
    indices = list(s.collected_pickup_indices)

    # Assert
    assert indices == [PickupIndex(1), PickupIndex(15)]


def test_add_pickup_to_state(state_game_data, empty_patches, generic_item_category):
    # Starting State
    s = state.State({}, (), 99, None, empty_patches, None, state_game_data)

    resource_a = ItemResourceInfo(1, "A", "A", 10, None)
    resource_b = ItemResourceInfo(2, "B", "B", 10, None)
    p = PickupEntry("B", 2, generic_item_category, generic_item_category,
                    progression=(
                        (resource_a, 1),
                        (resource_b, 1),
                    ))

    # Run
    state.add_pickup_to_state(s, p)
    state.add_pickup_to_state(s, p)

    # Assert
    assert s.resources == {
        resource_a: 1,
        resource_b: 1,
    }


def test_assign_pickup_to_starting_items(empty_patches, state_game_data, generic_item_category):
    # Setup

    starting = state.State({}, (), 99, None, empty_patches, None, state_game_data)

    resource_a = ItemResourceInfo(1, "A", "A", 10, None)
    resource_b = ItemResourceInfo(2, "B", "B", 10, None)
    p = PickupEntry("A", 2, generic_item_category, generic_item_category,
                    progression=(
                        (resource_a, 5),
                    ),
                    extra_resources=(
                        (state_game_data.resource_database.item_percentage, 1),
                    ),
                    unlocks_resource=True,
                    resource_lock=ResourceLock(resource_a, resource_a, resource_b),
                    )

    # Run
    final = starting.assign_pickup_to_starting_items(p)

    # Assert
    assert final.patches.starting_items == {resource_a: 5, resource_b: 0}
    assert final.resources == {resource_a: 5, resource_b: 0}


def test_state_with_pickup(state_game_data, empty_patches, generic_item_category):
    # Setup
    starting = state.State({}, (), 99, None, empty_patches, None, state_game_data)

    resource_a = ItemResourceInfo(1, "A", "A", 10, None)
    p = PickupEntry("A", 2, generic_item_category, generic_item_category,
                    progression=(
                        (resource_a, 1),
                    ))

    # Run
    final = state.state_with_pickup(starting, p)

    # Assert
    assert final.previous_state is starting
    assert final.resources == {resource_a: 1}



import dataclasses

from randovania.games.game import RandovaniaGame
from randovania.gui.generated.preset_patcher_energy_ui import Ui_PresetPatcherEnergy
from randovania.gui.lib import signal_handling
from randovania.gui.preset_settings.preset_tab import PresetTab
from randovania.interface_common.preset_editor import PresetEditor
from randovania.layout.preset import Preset
from randovania.layout.prime1.prime_configuration import PrimeConfiguration
from randovania.layout.prime2.echoes_configuration import EchoesConfiguration


class PresetPatcherEnergy(PresetTab, Ui_PresetPatcherEnergy):

    def __init__(self, editor: PresetEditor, game_enum: RandovaniaGame):
        super().__init__(editor)
        self.setupUi(self)
        self.game_enum = game_enum

        self.energy_tank_capacity_spin_box.valueChanged.connect(self._persist_tank_capacity)
        signal_handling.on_checked(self.dangerous_tank_check, self._persist_dangerous_tank)

        if self.game_enum == RandovaniaGame.METROID_PRIME_ECHOES:
            config_fields = {
                field.name: field
                for field in dataclasses.fields(EchoesConfiguration)
            }
            self.varia_suit_spin_box.setMinimum(config_fields["varia_suit_damage"].metadata["min"])
            self.varia_suit_spin_box.setMaximum(config_fields["varia_suit_damage"].metadata["max"])
            self.dark_suit_spin_box.setMinimum(config_fields["dark_suit_damage"].metadata["min"])
            self.dark_suit_spin_box.setMaximum(config_fields["dark_suit_damage"].metadata["max"])

            signal_handling.on_checked(self.safe_zone_logic_heal_check, self._persist_safe_zone_logic_heal)
            self.safe_zone_regen_spin.valueChanged.connect(self._persist_safe_zone_regen)
            self.varia_suit_spin_box.valueChanged.connect(self._persist_float("varia_suit_damage"))
            self.dark_suit_spin_box.valueChanged.connect(self._persist_float("dark_suit_damage"))
        else:
            self.dark_aether_box.setVisible(False)
            self.safe_zone_box.setVisible(False)

        if self.game_enum == RandovaniaGame.METROID_PRIME:
            config_fields = {
                field.name: field
                for field in dataclasses.fields(PrimeConfiguration)
            }
            self.heated_damage_spin.setMinimum(config_fields["heat_damage"].metadata["min"])
            self.heated_damage_spin.setMaximum(config_fields["heat_damage"].metadata["max"])

            signal_handling.on_checked(self.progressive_damage_reduction_check, self._persist_progressive_damage)
            signal_handling.on_checked(self.heated_damage_varia_check, self._persist_heat_protection_only_varia)
            self.heated_damage_spin.valueChanged.connect(self._persist_float("heat_damage"))

        else:
            self.progressive_damage_reduction_check.setVisible(False)
            self.heated_damage_box.setVisible(False)

    @property
    def uses_patches_tab(self) -> bool:
        return True

    def on_preset_changed(self, preset: Preset):
        config = preset.configuration
        self.energy_tank_capacity_spin_box.setValue(config.energy_per_tank)

        if self.game_enum == RandovaniaGame.METROID_PRIME_ECHOES:
            self.dangerous_tank_check.setChecked(config.dangerous_energy_tank)
            self.safe_zone_logic_heal_check.setChecked(config.safe_zone.fully_heal)
            self.safe_zone_regen_spin.setValue(config.safe_zone.heal_per_second)
            self.varia_suit_spin_box.setValue(config.varia_suit_damage)
            self.dark_suit_spin_box.setValue(config.dark_suit_damage)

        elif self.game_enum == RandovaniaGame.METROID_PRIME:
            self.progressive_damage_reduction_check.setChecked(config.progressive_damage_reduction)
            self.heated_damage_varia_check.setChecked(config.heat_protection_only_varia)
            self.heated_damage_spin.setValue(config.heat_damage)

    def _persist_tank_capacity(self):
        with self._editor as editor:
            editor.set_configuration_field("energy_per_tank", int(self.energy_tank_capacity_spin_box.value()))

    def _persist_safe_zone_regen(self):
        with self._editor as editor:
            safe_zone = dataclasses.replace(
                editor.configuration.safe_zone,
                heal_per_second=self.safe_zone_regen_spin.value()
            )
            editor.set_configuration_field("safe_zone", safe_zone)

    def _persist_safe_zone_logic_heal(self, checked: bool):
        with self._editor as editor:
            safe_zone = dataclasses.replace(
                editor.configuration.safe_zone,
                fully_heal=checked
            )
            editor.set_configuration_field("safe_zone", safe_zone)

    def _persist_progressive_damage(self, checked: bool):
        with self._editor as editor:
            editor.set_configuration_field("progressive_damage_reduction", checked)

    def _persist_heat_protection_only_varia(self, checked: bool):
        with self._editor as editor:
            editor.set_configuration_field("heat_protection_only_varia", checked)

    def _persist_dangerous_tank(self, checked: bool):
        with self._editor as editor:
            editor.set_configuration_field("dangerous_energy_tank", checked)




# -*- coding: utf-8 -*-

################################################################################
## Form generated from reading UI file 'preset_prime_patches.ui'
##
## Created by: Qt User Interface Compiler version 5.15.2
##
## WARNING! All changes made in this file will be lost when recompiling UI file!
################################################################################

from PySide2.QtCore import *
from PySide2.QtGui import *
from PySide2.QtWidgets import *


class Ui_PresetPrimePatches(object):
    def setupUi(self, PresetPrimePatches):
        if not PresetPrimePatches.objectName():
            PresetPrimePatches.setObjectName(u"PresetPrimePatches")
        PresetPrimePatches.resize(503, 675)
        self.centralWidget = QWidget(PresetPrimePatches)
        self.centralWidget.setObjectName(u"centralWidget")
        self.centralWidget.setMaximumSize(QSize(16777215, 16777215))
        self.verticalLayout = QVBoxLayout(self.centralWidget)
        self.verticalLayout.setSpacing(6)
        self.verticalLayout.setContentsMargins(11, 11, 11, 11)
        self.verticalLayout.setObjectName(u"verticalLayout")
        self.verticalLayout.setContentsMargins(0, 0, 0, 0)
        self.scroll_area = QScrollArea(self.centralWidget)
        self.scroll_area.setObjectName(u"scroll_area")
        self.scroll_area.setWidgetResizable(True)
        self.scroll_contents = QWidget()
        self.scroll_contents.setObjectName(u"scroll_contents")
        self.scroll_contents.setGeometry(QRect(0, 0, 487, 1189))
        self.scroll_layout = QVBoxLayout(self.scroll_contents)
        self.scroll_layout.setSpacing(6)
        self.scroll_layout.setContentsMargins(11, 11, 11, 11)
        self.scroll_layout.setObjectName(u"scroll_layout")
        self.scroll_layout.setContentsMargins(0, 2, 0, 0)
        self.top_spacer = QSpacerItem(20, 8, QSizePolicy.Minimum, QSizePolicy.Fixed)

        self.scroll_layout.addItem(self.top_spacer)

        self.warp_to_start_group = QGroupBox(self.scroll_contents)
        self.warp_to_start_group.setObjectName(u"warp_to_start_group")
        self.verticalLayout_6 = QVBoxLayout(self.warp_to_start_group)
        self.verticalLayout_6.setSpacing(6)
        self.verticalLayout_6.setContentsMargins(11, 11, 11, 11)
        self.verticalLayout_6.setObjectName(u"verticalLayout_6")
        self.warp_to_start_check = QCheckBox(self.warp_to_start_group)
        self.warp_to_start_check.setObjectName(u"warp_to_start_check")

        self.verticalLayout_6.addWidget(self.warp_to_start_check)

        self.warp_to_start_label = QLabel(self.warp_to_start_group)
        self.warp_to_start_label.setObjectName(u"warp_to_start_label")
        self.warp_to_start_label.setWordWrap(True)

        self.verticalLayout_6.addWidget(self.warp_to_start_label)


        self.scroll_layout.addWidget(self.warp_to_start_group)

        self.qol_group = QGroupBox(self.scroll_contents)
        self.qol_group.setObjectName(u"qol_group")
        self.verticalLayout_2 = QVBoxLayout(self.qol_group)
        self.verticalLayout_2.setSpacing(6)
        self.verticalLayout_2.setContentsMargins(11, 11, 11, 11)
        self.verticalLayout_2.setObjectName(u"verticalLayout_2")
        self.description_label = QLabel(self.qol_group)
        self.description_label.setObjectName(u"description_label")
        self.description_label.setWordWrap(True)
        self.description_label.setOpenExternalLinks(True)

        self.verticalLayout_2.addWidget(self.description_label)

        self.groupBox = QGroupBox(self.qol_group)
        self.groupBox.setObjectName(u"groupBox")
        self.verticalLayout_3 = QVBoxLayout(self.groupBox)
        self.verticalLayout_3.setSpacing(6)
        self.verticalLayout_3.setContentsMargins(11, 11, 11, 11)
        self.verticalLayout_3.setObjectName(u"verticalLayout_3")
        self.main_plaza_door_check = QCheckBox(self.groupBox)
        self.main_plaza_door_check.setObjectName(u"main_plaza_door_check")

        self.verticalLayout_3.addWidget(self.main_plaza_door_check)

        self.main_plaza_door_label = QLabel(self.groupBox)
        self.main_plaza_door_label.setObjectName(u"main_plaza_door_label")

        self.verticalLayout_3.addWidget(self.main_plaza_door_label)

        self.backwards_frigate_check = QCheckBox(self.groupBox)
        self.backwards_frigate_check.setObjectName(u"backwards_frigate_check")

        self.verticalLayout_3.addWidget(self.backwards_frigate_check)

        self.backwards_frigate_label = QLabel(self.groupBox)
        self.backwards_frigate_label.setObjectName(u"backwards_frigate_label")

        self.verticalLayout_3.addWidget(self.backwards_frigate_label)

        self.backwards_labs_check = QCheckBox(self.groupBox)
        self.backwards_labs_check.setObjectName(u"backwards_labs_check")

        self.verticalLayout_3.addWidget(self.backwards_labs_check)

        self.backwards_labs_label = QLabel(self.groupBox)
        self.backwards_labs_label.setObjectName(u"backwards_labs_label")

        self.verticalLayout_3.addWidget(self.backwards_labs_label)

        self.backwards_upper_mines_check = QCheckBox(self.groupBox)
        self.backwards_upper_mines_check.setObjectName(u"backwards_upper_mines_check")

        self.verticalLayout_3.addWidget(self.backwards_upper_mines_check)

        self.backwards_upper_mines_label = QLabel(self.groupBox)
        self.backwards_upper_mines_label.setObjectName(u"backwards_upper_mines_label")

        self.verticalLayout_3.addWidget(self.backwards_upper_mines_label)

        self.backwards_lower_mines_check = QCheckBox(self.groupBox)
        self.backwards_lower_mines_check.setObjectName(u"backwards_lower_mines_check")

        self.verticalLayout_3.addWidget(self.backwards_lower_mines_check)

        self.backwards_lower_mines_label = QLabel(self.groupBox)
        self.backwards_lower_mines_label.setObjectName(u"backwards_lower_mines_label")

        self.verticalLayout_3.addWidget(self.backwards_lower_mines_label)

        self.phazon_elite_without_dynamo_check = QCheckBox(self.groupBox)
        self.phazon_elite_without_dynamo_check.setObjectName(u"phazon_elite_without_dynamo_check")

        self.verticalLayout_3.addWidget(self.phazon_elite_without_dynamo_check)

        self.phazon_elite_without_dynamo_label = QLabel(self.groupBox)
        self.phazon_elite_without_dynamo_label.setObjectName(u"phazon_elite_without_dynamo_label")

        self.verticalLayout_3.addWidget(self.phazon_elite_without_dynamo_label)


        self.verticalLayout_2.addWidget(self.groupBox)

        self.qol_game_breaking_check = QCheckBox(self.qol_group)
        self.qol_game_breaking_check.setObjectName(u"qol_game_breaking_check")
        self.qol_game_breaking_check.setEnabled(True)

        self.verticalLayout_2.addWidget(self.qol_game_breaking_check)

        self.qol_game_breaking_label = QLabel(self.qol_group)
        self.qol_game_breaking_label.setObjectName(u"qol_game_breaking_label")
        sizePolicy = QSizePolicy(QSizePolicy.Preferred, QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.qol_game_breaking_label.sizePolicy().hasHeightForWidth())
        self.qol_game_breaking_label.setSizePolicy(sizePolicy)
        self.qol_game_breaking_label.setMaximumSize(QSize(16777215, 60))
        self.qol_game_breaking_label.setAlignment(Qt.AlignLeading|Qt.AlignLeft|Qt.AlignTop)
        self.qol_game_breaking_label.setWordWrap(True)
        self.qol_game_breaking_label.setOpenExternalLinks(True)

        self.verticalLayout_2.addWidget(self.qol_game_breaking_label)

        self.qol_cosmetic_check = QCheckBox(self.qol_group)
        self.qol_cosmetic_check.setObjectName(u"qol_cosmetic_check")
        self.qol_cosmetic_check.setEnabled(False)
        self.qol_cosmetic_check.setChecked(True)

        self.verticalLayout_2.addWidget(self.qol_cosmetic_check)

        self.qol_cosmetic_label = QLabel(self.qol_group)
        self.qol_cosmetic_label.setObjectName(u"qol_cosmetic_label")
        self.qol_cosmetic_label.setWordWrap(True)

        self.verticalLayout_2.addWidget(self.qol_cosmetic_label)

        self.qol_pickup_scans_check = QCheckBox(self.qol_group)
        self.qol_pickup_scans_check.setObjectName(u"qol_pickup_scans_check")

        self.verticalLayout_2.addWidget(self.qol_pickup_scans_check)

        self.qol_pickup_scans_label = QLabel(self.qol_group)
        self.qol_pickup_scans_label.setObjectName(u"qol_pickup_scans_label")
        self.qol_pickup_scans_label.setWordWrap(True)

        self.verticalLayout_2.addWidget(self.qol_pickup_scans_label)


        self.scroll_layout.addWidget(self.qol_group)

        self.cutscene_group = QGroupBox(self.scroll_contents)
        self.cutscene_group.setObjectName(u"cutscene_group")
        self.verticalLayout_5 = QVBoxLayout(self.cutscene_group)
        self.verticalLayout_5.setSpacing(6)
        self.verticalLayout_5.setContentsMargins(11, 11, 11, 11)
        self.verticalLayout_5.setObjectName(u"verticalLayout_5")
        self.cutscene_combo = QComboBox(self.cutscene_group)
        self.cutscene_combo.addItem("")
        self.cutscene_combo.addItem("")
        self.cutscene_combo.addItem("")
        self.cutscene_combo.addItem("")
        self.cutscene_combo.setObjectName(u"cutscene_combo")

        self.verticalLayout_5.addWidget(self.cutscene_combo)

        self.cutscene_label = QLabel(self.cutscene_group)
        self.cutscene_label.setObjectName(u"cutscene_label")
        self.cutscene_label.setWordWrap(True)

        self.verticalLayout_5.addWidget(self.cutscene_label)


        self.scroll_layout.addWidget(self.cutscene_group)

        self.small_samus_group = QGroupBox(self.scroll_contents)
        self.small_samus_group.setObjectName(u"small_samus_group")
        self.verticalLayout_4 = QVBoxLayout(self.small_samus_group)
        self.verticalLayout_4.setSpacing(6)
        self.verticalLayout_4.setContentsMargins(11, 11, 11, 11)
        self.verticalLayout_4.setObjectName(u"verticalLayout_4")
        self.small_samus_check = QCheckBox(self.small_samus_group)
        self.small_samus_check.setObjectName(u"small_samus_check")

        self.verticalLayout_4.addWidget(self.small_samus_check)

        self.small_samus_label = QLabel(self.small_samus_group)
        self.small_samus_label.setObjectName(u"small_samus_label")
        self.small_samus_label.setWordWrap(True)

        self.verticalLayout_4.addWidget(self.small_samus_label)


        self.scroll_layout.addWidget(self.small_samus_group)

        self.scroll_area.setWidget(self.scroll_contents)

        self.verticalLayout.addWidget(self.scroll_area)

        PresetPrimePatches.setCentralWidget(self.centralWidget)

        self.retranslateUi(PresetPrimePatches)

        QMetaObject.connectSlotsByName(PresetPrimePatches)
    # setupUi

    def retranslateUi(self, PresetPrimePatches):
        PresetPrimePatches.setWindowTitle(QCoreApplication.translate("PresetPrimePatches", u"Other", None))
        self.warp_to_start_group.setTitle(QCoreApplication.translate("PresetPrimePatches", u"Warp to Start", None))
        self.warp_to_start_check.setText(QCoreApplication.translate("PresetPrimePatches", u"Add warping to starting location from save stations", None))
        self.warp_to_start_label.setText(QCoreApplication.translate("PresetPrimePatches", u"<html><head/><body><p>Refusing to save at any Save Station while holding L+R will warp you to the starting location (by default, Samus' ship in Landing Site).</p></body></html>", None))
        self.qol_group.setTitle(QCoreApplication.translate("PresetPrimePatches", u"Quality of life", None))
        self.description_label.setText(QCoreApplication.translate("PresetPrimePatches", u"<html><head/><body><p>In order to improve the game experience, many quality of life changes are provided separated in 5 different groups.</p><p>For in-depth list of changes, see <a href=\"https://github.com/toasterparty/randomprime/blob/randovania/doc/quality_of_life.md\"><span style=\" text-decoration: underline; color:#0000ff;\">Quality of Life</span></a>.</p></body></html>", None))
        self.groupBox.setTitle(QCoreApplication.translate("PresetPrimePatches", u"Logical changes", None))
        self.main_plaza_door_check.setText(QCoreApplication.translate("PresetPrimePatches", u"Main Plaza Vault Door", None))
        self.main_plaza_door_label.setText(QCoreApplication.translate("PresetPrimePatches", u"Allow access to Vault from Main Plaza", None))
        self.backwards_frigate_check.setText(QCoreApplication.translate("PresetPrimePatches", u"Backwards Frigate", None))
        self.backwards_frigate_label.setText(QCoreApplication.translate("PresetPrimePatches", u"Power door in Main Ventilation Shaft Section B when approached from behind", None))
        self.backwards_labs_check.setText(QCoreApplication.translate("PresetPrimePatches", u"Backwards Labs", None))
        self.backwards_labs_label.setText(QCoreApplication.translate("PresetPrimePatches", u"Scan through barrier of Research Lab Hydra when approached from deep labs", None))
        self.backwards_upper_mines_check.setText(QCoreApplication.translate("PresetPrimePatches", u"Backwards Upper Mines", None))
        self.backwards_upper_mines_label.setText(QCoreApplication.translate("PresetPrimePatches", u"Disable Main Quarry barrier automatically when approached from Mine Security Station", None))
        self.backwards_lower_mines_check.setText(QCoreApplication.translate("PresetPrimePatches", u"Backwards Lower Mines", None))
        self.backwards_lower_mines_label.setText(QCoreApplication.translate("PresetPrimePatches", u"Remove PCA locks and allow passing through lower mines scan barriers from the back", None))
        self.phazon_elite_without_dynamo_check.setText(QCoreApplication.translate("PresetPrimePatches", u"Phazon Elite without Dynamo", None))
        self.phazon_elite_without_dynamo_label.setText(QCoreApplication.translate("PresetPrimePatches", u"Removes the Central Dynamo item requirement for activating the Phazon Elite boss fight", None))
        self.qol_game_breaking_check.setText(QCoreApplication.translate("PresetPrimePatches", u"Game Breaking", None))
        self.qol_game_breaking_label.setText(QCoreApplication.translate("PresetPrimePatches", u"<html><head/><body><p>Fixes softlocks and wrong room spawn points. Reverts changes made in later game versions.<br/>Logic always assume this is enabled.</p></body></html>", None))
        self.qol_cosmetic_check.setText(QCoreApplication.translate("PresetPrimePatches", u"Cosmetic", None))
        self.qol_cosmetic_label.setText(QCoreApplication.translate("PresetPrimePatches", u"<html><head/><body><p>Remove or change things that only affect real time, not in-game time.<br/>This group is controlled by an user preference and changing it does not affect the permalink.</p></body></html>", None))
        self.qol_pickup_scans_check.setText(QCoreApplication.translate("PresetPrimePatches", u"Pickup Scans", None))
        self.qol_pickup_scans_label.setText(QCoreApplication.translate("PresetPrimePatches", u"Enable this option to patch various item locations so that they can be scanned. Only locations whose model is visible in combat visor are affected. Note that the new scan may be near the item, not necessarily on top of it.", None))
        self.cutscene_group.setTitle(QCoreApplication.translate("PresetPrimePatches", u"Cutscene removal mode", None))
        self.cutscene_combo.setItemText(0, QCoreApplication.translate("PresetPrimePatches", u"Original", None))
        self.cutscene_combo.setItemText(1, QCoreApplication.translate("PresetPrimePatches", u"Competitive", None))
        self.cutscene_combo.setItemText(2, QCoreApplication.translate("PresetPrimePatches", u"Minor", None))
        self.cutscene_combo.setItemText(3, QCoreApplication.translate("PresetPrimePatches", u"Major", None))

        self.cutscene_label.setText(QCoreApplication.translate("PresetPrimePatches", u"<html><head/><body><p><span style=\" font-weight:600;\">Original</span>: No changes to the cutscenes are made.</p><p><span style=\" font-weight:600;\">Competitive</span>: Similar to minor, but leaves a few rooms alone where skipping cutscenes would be inappropriate for races.</p><p><span style=\" font-weight:600;\">Minor</span>: Removes cutscenes that don't affect the game too much when removed.</p><p><span style=\" font-weight:600;\">Major</span>: Allows you to continue playing the game while cutscenes happen.</p></body></html>", None))
        self.small_samus_group.setTitle(QCoreApplication.translate("PresetPrimePatches", u"Small Samus", None))
        self.small_samus_check.setText(QCoreApplication.translate("PresetPrimePatches", u"Enable Small Samus", None))
        self.small_samus_label.setText(QCoreApplication.translate("PresetPrimePatches", u"<html><head/><body><p>In this mode, Samus is scaled to 30% of her normal size. Expect to be able to fit in many places you weren't before, but also fall down in a lot of holes you didn't before.</p><p><span style=\" font-weight:600;\">Warning:</span> This option has been known to induce motion sickness and nausea in some players. If you are concerned you may suffer from these symptoms, it may be best to avoid playing with this setting enabled.</p></body></html>", None))
    # retranslateUi




import copy
import functools
import json
import shutil
from pathlib import Path
from typing import Optional, List

from randovania import get_data_path
from randovania.games.patcher import Patcher
from randovania.games.patchers import claris_randomizer, claris_patcher_file
from randovania.games.patchers.gamecube import banner_patcher, iso_packager
from randovania.interface_common import game_workdir
from randovania.interface_common.players_configuration import PlayersConfiguration
from randovania.layout.layout_description import LayoutDescription
from randovania.layout.prime2.echoes_cosmetic_patches import EchoesCosmeticPatches
from randovania.lib import status_update_lib


class ClarisPatcher(Patcher):
    _busy: bool = False

    @property
    def is_busy(self) -> bool:
        """
        Checks if the patcher is busy right now
        """
        return self._busy

    @property
    def export_can_be_aborted(self) -> bool:
        """
        Checks if patch_game can be aborted
        """
        return True

    @property
    def uses_input_file_directly(self) -> bool:
        """
        Does this patcher uses the input file directly?
        """
        return False

    def has_internal_copy(self, internal_copies_path: Path) -> bool:
        result = game_workdir.discover_game(internal_copies_path.joinpath("prime2", "contents"))
        if result is not None:
            game_id, _ = result
            if game_id.startswith("G2M"):
                return True
        return False

    def delete_internal_copy(self, internal_copies_path: Path):
        internal_copies_path = internal_copies_path.joinpath("prime2")
        if internal_copies_path.exists():
            shutil.rmtree(internal_copies_path)

    def default_output_file(self, seed_hash: str) -> str:
        return "Echoes Randomizer - {}".format(seed_hash)

    @property
    def valid_input_file_types(self) -> List[str]:
        return ["iso"]

    @property
    def valid_output_file_types(self) -> List[str]:
        return ["iso"]

    def create_patch_data(self, description: LayoutDescription, players_config: PlayersConfiguration,
                          cosmetic_patches: EchoesCosmeticPatches):
        return claris_patcher_file.create_patcher_file(description, players_config, cosmetic_patches)

    def patch_game(self, input_file: Optional[Path], output_file: Path, patch_data: dict,
                   internal_copies_path: Path, progress_update: status_update_lib.ProgressUpdateCallable):
        updaters = status_update_lib.split_progress_update(progress_update, 4)

        contents_files_path = internal_copies_path.joinpath("prime2", "contents")
        backup_files_path = internal_copies_path.joinpath("prime2", "vanilla")

        if input_file is not None:
            unpack_updaters = status_update_lib.split_progress_update(updaters[0], 2)
            self.delete_internal_copy(internal_copies_path)
            iso_packager.unpack_iso(
                iso=input_file,
                game_files_path=contents_files_path,
                progress_update=unpack_updaters[0],
            )
            claris_randomizer.create_pak_backups(
                contents_files_path,
                backup_files_path,
                unpack_updaters[1]
            )
        else:
            try:
                claris_randomizer.restore_pak_backups(
                    contents_files_path,
                    backup_files_path,
                    updaters[0]
                )
            except FileNotFoundError:
                raise RuntimeError(
                    "Your internal copy is missing files.\nPlease press 'Delete internal copy' and select "
                    "a clean game ISO.")

        # Apply patcher
        banner_patcher.patch_game_name_and_id(
            contents_files_path,
            "Metroid Prime 2: Randomizer - {}".format(patch_data["shareable_hash"]),
            patch_data["publisher_id"]
        )
        randomizer_data = copy.deepcopy(decode_randomizer_data())

        if patch_data.pop("convert_other_game_assets", False):
            from randovania.games.prime import asset_conversion
            asset_conversion.convert_prime1_pickups(contents_files_path, randomizer_data, updaters[1])

        claris_patcher_file.adjust_model_name(patch_data, randomizer_data)
        claris_randomizer.apply_patcher_file(
            contents_files_path,
            patch_data,
            randomizer_data,
            updaters[2])

        # Pack ISO
        iso_packager.pack_iso(
            iso=output_file,
            game_files_path=contents_files_path,
            progress_update=updaters[3],
        )


@functools.lru_cache()
def decode_randomizer_data() -> dict:
    randomizer_data_path = get_data_path().joinpath("ClarisPrimeRandomizer", "RandomizerData.json")

    with randomizer_data_path.open() as randomizer_data_file:
        return json.load(randomizer_data_file)



import json
import logging
import shutil
from pathlib import Path
from typing import Callable, List, Union

from randovania import get_data_path
from randovania.games.patchers import csharp_subprocess
from randovania.games.patchers.exceptions import ExportFailure
from randovania.games.prime import echoes_dol_patcher
from randovania.interface_common.game_workdir import validate_game_files_path
from randovania.lib import status_update_lib
from randovania.lib.status_update_lib import ProgressUpdateCallable

try:
    from asyncio.exceptions import IncompleteReadError
except ImportError:
    from asyncio.streams import IncompleteReadError

CURRENT_PATCH_VERSION = 2
logger = logging.getLogger(__name__)


def _patch_version_file(game_root: Path) -> Path:
    return game_root.joinpath("randovania_patch_version.txt")


def get_patch_version(game_root: Path) -> int:
    file = _patch_version_file(game_root)
    if file.exists():
        return int(file.read_text("utf-8"))
    else:
        return 0


def write_patch_version(game_root: Path, version: int):
    _patch_version_file(game_root).write_text(str(version))


def _get_randomizer_folder() -> Path:
    return get_data_path().joinpath("ClarisPrimeRandomizer")


def _get_randomizer_path() -> Path:
    return _get_randomizer_folder().joinpath("Randomizer.exe")


def _get_custom_data_path() -> Path:
    return _get_randomizer_folder().joinpath("CustomRandomizerData.json")


def _get_menu_mod_path() -> Path:
    return get_data_path().joinpath("ClarisEchoesMenu", "EchoesMenu.exe")


def _run_with_args(args: List[Union[str, Path]],
                   input_data: str,
                   finish_string: str,
                   status_update: Callable[[str], None]):
    finished_updates = False

    new_args = [str(arg) for arg in args]
    logger.info("Invoking external tool with: %s", new_args)
    all_lines = []

    def read_callback(line: str):
        nonlocal finished_updates
        logger.info(line)
        all_lines.append(line)
        if not finished_updates:
            status_update(line)
            finished_updates = line == finish_string

    csharp_subprocess.process_command(new_args, input_data, read_callback)

    if not finished_updates:
        raise ExportFailure(
            "External tool did not send '{}'.".format(finish_string),
            "\n".join(all_lines),
        )


def _base_args(game_root: Path,
               ) -> List[Union[str, Path]]:
    game_files = game_root / "files"
    validate_game_files_path(game_files)

    return [
        _get_randomizer_path(),
        game_root,
        # "-test",
        "-data:" + str(_get_custom_data_path()),
    ]


_ECHOES_PAKS = tuple(
    [
        "MiscData.pak",
        "FrontEnd.pak",
        "LogBook.pak",
    ]
    + ["Metroid{}.pak".format(i) for i in range(1, 6)])


def restore_pak_backups(
        game_root: Path,
        backup_files_path: Path,
        progress_update: ProgressUpdateCallable,
):
    """
    Ensures the given game_root has unmodified paks.
    :param game_root:
    :param backup_files_path:
    :param progress_update:
    :return:
    """
    pak_folder = backup_files_path.joinpath("paks")
    files_folder = game_root.joinpath("files")
    for i, pak in enumerate(_ECHOES_PAKS):
        progress_update("Restoring {} from backup".format(pak), i / len(_ECHOES_PAKS))
        shutil.copy(pak_folder.joinpath(pak), files_folder.joinpath(pak))

    files_folder.joinpath("menu_mod.txt").unlink(missing_ok=True)


def create_pak_backups(
        game_root: Path,
        backup_files_path: Path,
        progress_update: ProgressUpdateCallable,
):
    pak_folder = backup_files_path.joinpath("paks")
    pak_folder.mkdir(parents=True, exist_ok=True)

    files_folder = game_root.joinpath("files")
    for i, pak in enumerate(_ECHOES_PAKS):
        progress_update("Backing up {}".format(pak), i / len(_ECHOES_PAKS))
        shutil.copy(files_folder.joinpath(pak), pak_folder.joinpath(pak))


def _add_menu_mod_to_files(
        game_root: Path,
        status_update: Callable[[str], None],
):
    files_folder = game_root.joinpath("files")
    _run_with_args(
        [_get_menu_mod_path(), files_folder],
        "",
        "Done!",
        status_update
    )
    files_folder.joinpath("menu_mod.txt").write_bytes(b"")


def apply_patcher_file(game_root: Path,
                       patcher_data: dict,
                       randomizer_data: dict,
                       progress_update: ProgressUpdateCallable,
                       ):
    """
    Applies the modifications listed in the given patcher_data to the game in game_root.
    :param game_root:
    :param patcher_data:
    :param randomizer_data: The RandomizerData.json contents to use.
    :param progress_update:
    :return:
    """
    menu_mod = patcher_data["menu_mod"]

    status_update = status_update_lib.create_progress_update_from_successive_messages(
        progress_update, 200 if menu_mod else 100)

    last_version = get_patch_version(game_root)
    if last_version > CURRENT_PATCH_VERSION:
        raise ExportFailure(f"Game at {game_root} was last patched with version {last_version}, "
                            f"which is above supported version {CURRENT_PATCH_VERSION}. "
                            f"\nPlease press 'Delete internal copy'.", None)

    with _get_custom_data_path().open("w") as data_file:
        json.dump(randomizer_data, data_file, indent=4)
    _run_with_args(_base_args(game_root),
                   json.dumps(patcher_data),
                   "Randomized!",
                   status_update)
    echoes_dol_patcher.apply_patches(game_root,
                                     echoes_dol_patcher.EchoesDolPatchesData.from_json(patcher_data["dol_patches"]))
    write_patch_version(game_root, CURRENT_PATCH_VERSION)

    if menu_mod:
        _add_menu_mod_to_files(game_root, status_update)



import base64
import binascii
import json
import operator
from dataclasses import dataclass
from typing import Iterator, Tuple, Dict, Iterable

import bitstruct

from randovania.bitpacking import bitpacking
from randovania.bitpacking.bitpacking import BitPackDecoder, BitPackValue, single_byte_hash
from randovania.interface_common.preset_manager import PresetManager
from randovania.layout.preset import Preset

_PERMALINK_MAX_VERSION = 16
_PERMALINK_MAX_SEED = 2 ** 31
_PERMALINK_PLAYER_COUNT_LIMITS = (2, 256)


def _dictionary_byte_hash(data: dict) -> int:
    return single_byte_hash(json.dumps(data, separators=(',', ':')).encode("UTF-8"))


def rotate_bytes(data: Iterable[int], rotation: int, per_byte_adjustment: int,
                 inverse: bool = False) -> Iterator[int]:
    """
    Rotates the elements in data, in a reversible operation.
    :param data: The byte values to rotate. Should be ints in the [0, 255] range.
    :param rotation: By how much each item should be rotated, mod 256.
    :param per_byte_adjustment: Increment the rotation by this after each byte, mod 256.
    :param inverse: If True, it performs the inverse operation.
    :return:
    """
    if inverse:
        op = operator.sub
    else:
        op = operator.add
    for b in data:
        yield op(b, rotation) % 256
        rotation = (rotation + per_byte_adjustment) % 256


@dataclass(frozen=True)
class Permalink(BitPackValue):
    seed_number: int
    spoiler: bool
    presets: Dict[int, Preset]

    def __post_init__(self):
        if self.seed_number is None:
            raise ValueError("Missing seed number")
        if not (0 <= self.seed_number < _PERMALINK_MAX_SEED):
            raise ValueError("Invalid seed number: {}".format(self.seed_number))
        object.__setattr__(self, "__cached_as_bytes", None)

    @classmethod
    def current_version(cls) -> int:
        # When this reaches _PERMALINK_MAX_VERSION, we need to change how we decode to avoid breaking version detection
        # for previous Randovania versions
        return 12

    def bit_pack_encode(self, metadata) -> Iterator[Tuple[int, int]]:
        yield self.current_version(), _PERMALINK_MAX_VERSION
        yield self.seed_number, _PERMALINK_MAX_SEED
        yield from bitpacking.encode_bool(self.spoiler)
        yield from bitpacking.encode_int_with_limits(self.player_count, _PERMALINK_PLAYER_COUNT_LIMITS)

        manager = PresetManager(None)

        previous_unique_presets = []
        for preset in self.presets.values():
            already_encoded_preset = preset in previous_unique_presets
            yield from bitpacking.encode_bool(already_encoded_preset)
            if already_encoded_preset:
                yield from bitpacking.pack_array_element(preset, previous_unique_presets)
            else:
                previous_unique_presets.append(preset)
                yield from preset.bit_pack_encode({"manager": manager})

    @classmethod
    def bit_pack_unpack(cls, decoder: BitPackDecoder, metadata) -> "Permalink":
        version, seed_number = decoder.decode(_PERMALINK_MAX_VERSION, _PERMALINK_MAX_SEED)
        cls._raise_if_different_version(version)

        spoiler = bitpacking.decode_bool(decoder)
        player_count = bitpacking.decode_int_with_limits(decoder, _PERMALINK_PLAYER_COUNT_LIMITS)
        manager = PresetManager(None)

        previous_unique_presets = []
        presets = {}

        for index in range(player_count):
            in_previous_presets = bitpacking.decode_bool(decoder)
            if in_previous_presets:
                preset = decoder.decode_element(previous_unique_presets)
            else:
                preset = Preset.bit_pack_unpack(decoder, {"manager": manager})
                previous_unique_presets.append(preset)
            presets[index] = preset

        return Permalink(seed_number, spoiler, presets)

    @classmethod
    def _raise_if_different_version(cls, version: int):
        if version != cls.current_version():
            raise ValueError("Given permalink has version {}, but this Randovania "
                             "support only permalink of version {}.".format(version, cls.current_version()))

    @classmethod
    def validate_version(cls, b: bytes):
        version = BitPackDecoder(b).peek(_PERMALINK_MAX_VERSION)[0]
        cls._raise_if_different_version(version)

    @property
    def as_bytes(self) -> bytes:
        cached_result = object.__getattribute__(self, "__cached_as_bytes")
        if cached_result is not None:
            return cached_result

        encoded = bitpacking.pack_value(self)
        # Add extra bytes so the base64 encoding never uses == at the end
        encoded += b"\x00" * (3 - (len(encoded) + 1) % 3)

        # Rotate bytes, so the slightest change causes a cascading effect.
        # But skip the first byte so the version check can be done early in decoding.
        byte_hash = single_byte_hash(encoded)
        new_bytes = [encoded[0]]
        new_bytes.extend(rotate_bytes(encoded[1:], byte_hash, byte_hash, inverse=False))

        # Append the hash, so the rotation can be reversed and the checksum verified
        new_bytes.append(byte_hash)

        result = bytes(new_bytes)
        object.__setattr__(self, "__cached_as_bytes", result)
        return result

    @classmethod
    def from_bytes(cls, b: bytes) -> "Permalink":
        Permalink.validate_version(b)

        byte_hash = b[-1]
        new_bytes = [b[0]]
        new_bytes.extend(rotate_bytes(b[1:-1], byte_hash, byte_hash, inverse=True))

        decoded = bytes(new_bytes)
        if single_byte_hash(decoded) != byte_hash:
            raise ValueError("Incorrect checksum")

        decoder = BitPackDecoder(decoded)
        return Permalink.bit_pack_unpack(decoder, {})

    @property
    def as_base64_str(self) -> str:
        try:
            return base64.b64encode(self.as_bytes, altchars=b'-_').decode("utf-8")
        except ValueError as e:
            return "Unable to create Permalink: {}".format(e)

    @classmethod
    def from_str(cls, param: str) -> "Permalink":
        try:
            b = base64.b64decode(param.encode("utf-8"), altchars=b'-_', validate=True)
            if len(b) < 2:
                raise ValueError("Data too small")

            return cls.from_bytes(b)

        except (binascii.Error, bitstruct.Error) as e:
            raise ValueError("Unable to base64 decode '{permalink}': {error}".format(
                permalink=param,
                error=e,
            ))

    @property
    def player_count(self) -> int:
        return len(self.presets)

    def get_preset(self, index: int) -> Preset:
        return self.presets[index]



from dataclasses import dataclass
from typing import Dict


@dataclass(frozen=True)
class PlayersConfiguration:
    player_index: int
    player_names: Dict[int, str]

    @property
    def is_multiworld(self) -> int:
        return len(self.player_names) > 1



import asyncio
import datetime
import os
import pprint
import subprocess
import typing

import aiohttp

from randovania import VERSION


async def post_to_discord():
    git_result = subprocess.run(["git", "show"], check=True, stdout=subprocess.PIPE, text=True).stdout.split("\n")
    commit_hash = git_result[0].split()[1]

    message = ""
    for line in git_result[4:]:
        if line.startswith("diff --git"):
            break
        else:
            message += line
            message += "\n"

    run_id = os.environ["GITHUB_RUN_ID"]
    org_name = "randovania"
    repo_name = "randovania"

    async with aiohttp.ClientSession() as session:
        session.headers['Authorization'] = f"Bearer {os.environ['GITHUB_TOKEN']}"

        run_details_url = f"https://api.github.com/repos/{org_name}/{repo_name}/actions/runs/{run_id}"
        print(f"> Get run details: {run_details_url}")
        async with session.get(run_details_url, raise_for_status=True) as response:
            run_details = await response.json()
            pprint.pprint(run_details)
            check_suite_id = run_details["check_suite_id"]

        artifacts_url = f"https://api.github.com/repos/{org_name}/{repo_name}/actions/runs/{run_id}/artifacts"
        print(f"> Get artifact details: {artifacts_url}")
        async with session.get(artifacts_url, raise_for_status=True) as response:
            artifacts = await response.json()
            pprint.pprint(artifacts)

    fields = [
        {
            "name": typing.cast(str, artifact["name"]).replace("Executable", "").strip(),
            "value": f"[Download](https://github.com/{org_name}/{repo_name}/suites/{check_suite_id}/artifacts/{artifact['id']})",
            "inline": True
        }
        for artifact in artifacts["artifacts"]
    ]
    fields.append(
        {
            "name": "Downloads",
            "value": f"[Link](https://github.com/{org_name}/{repo_name}/actions/runs/{run_id})",
            "inline": True
        }
    )

    webhook_data = {
        "embeds": [{
            "color": 0x2ecc71,
            "title": f"Randovania {VERSION}",
            "url": f"https://github.com/randovania/randovania/commit/{commit_hash}",
            "description": message.strip(),
            "fields": fields,
            "timestamp": datetime.datetime.now().isoformat()
        }]
    }
    webhook_url = os.environ["DISCORD_WEBHOOK"]
    async with aiohttp.ClientSession() as session:
        async with session.post(webhook_url, json=webhook_data, raise_for_status=True) as response:
            print(await response.text())


if __name__ == '__main__':
    asyncio.run(post_to_discord())



"""Classes that describes the raw data of a game world."""
import copy
import dataclasses
import typing
from typing import Iterator, FrozenSet, Dict, Optional, List, Tuple

from randovania.game_description.game_patches import GamePatches
from randovania.game_description.requirements import SatisfiableRequirements, Requirement
from randovania.game_description.resources.resource_database import ResourceDatabase
from randovania.game_description.resources.resource_info import ResourceInfo, ResourceGainTuple, CurrentResources
from randovania.game_description.resources.simple_resource_info import SimpleResourceInfo
from randovania.game_description.world.area import Area
from randovania.game_description.world.area_location import AreaLocation
from randovania.game_description.world.dock import DockWeaknessDatabase, DockWeakness
from randovania.game_description.world.node import TeleporterNode
from randovania.game_description.world.teleporter import Teleporter
from randovania.game_description.world.world_list import WorldList
from randovania.games.game import RandovaniaGame


def _calculate_dangerous_resources_in_db(
        db: DockWeaknessDatabase,
        database: ResourceDatabase,
) -> Iterator[ResourceInfo]:
    for list_by_type in db:
        for dock_weakness in typing.cast(List[DockWeakness], list_by_type):
            yield from dock_weakness.requirement.as_set(database).dangerous_resources


def _calculate_dangerous_resources_in_areas(
        areas: Iterator[Area],
        database: ResourceDatabase,
) -> Iterator[ResourceInfo]:
    for area in areas:
        for node in area.nodes:
            for requirement in area.connections[node].values():
                yield from requirement.as_set(database).dangerous_resources


@dataclasses.dataclass(frozen=True)
class IndexWithReason:
    index: int
    reason: Optional[str]


@dataclasses.dataclass(frozen=True)
class MinimalLogicData:
    items_to_exclude: List[IndexWithReason]
    custom_item_amount: Dict[int, int]
    events_to_exclude: List[IndexWithReason]


class GameDescription:
    game: RandovaniaGame
    dock_weakness_database: DockWeaknessDatabase

    resource_database: ResourceDatabase
    victory_condition: Requirement
    starting_location: AreaLocation
    initial_states: Dict[str, ResourceGainTuple]
    minimal_logic: Optional[MinimalLogicData]
    _dangerous_resources: Optional[FrozenSet[ResourceInfo]] = None
    world_list: WorldList
    mutable: bool = False

    def __deepcopy__(self, memodict):
        new_game = GameDescription(
            game=self.game,
            resource_database=self.resource_database,
            dock_weakness_database=self.dock_weakness_database,
            world_list=copy.deepcopy(self.world_list, memodict),
            victory_condition=self.victory_condition,
            starting_location=self.starting_location,
            initial_states=copy.copy(self.initial_states),
            minimal_logic=self.minimal_logic,
        )
        new_game._dangerous_resources = self._dangerous_resources
        return new_game

    def __init__(self,
                 game: RandovaniaGame,
                 dock_weakness_database: DockWeaknessDatabase,

                 resource_database: ResourceDatabase,
                 victory_condition: Requirement,
                 starting_location: AreaLocation,
                 initial_states: Dict[str, ResourceGainTuple],
                 minimal_logic: Optional[MinimalLogicData],
                 world_list: WorldList,
                 ):
        self.game = game
        self.dock_weakness_database = dock_weakness_database

        self.resource_database = resource_database
        self.victory_condition = victory_condition
        self.starting_location = starting_location
        self.initial_states = initial_states
        self.minimal_logic = minimal_logic
        self.world_list = world_list

    def patch_requirements(self, resources, damage_multiplier: float):
        if not self.mutable:
            raise ValueError("self is not mutable")

        self.world_list.patch_requirements(resources, damage_multiplier, self.resource_database)
        self._dangerous_resources = None

    def create_game_patches(self) -> GamePatches:
        elevator_connection: Dict[Teleporter, AreaLocation] = {
            node.teleporter: node.default_connection

            for node in self.world_list.all_nodes
            if isinstance(node, TeleporterNode) and node.editable
        }

        return GamePatches(None, {}, elevator_connection, {}, {}, {}, {}, self.starting_location, {})

    @property
    def dangerous_resources(self) -> FrozenSet[ResourceInfo]:
        if self._dangerous_resources is None:
            self._dangerous_resources = frozenset(
                _calculate_dangerous_resources_in_areas(self.world_list.all_areas, self.resource_database)) | frozenset(
                _calculate_dangerous_resources_in_db(self.dock_weakness_database, self.resource_database))
        return self._dangerous_resources

    def make_mutable_copy(self) -> "GameDescription":
        result = copy.deepcopy(self)
        result.mutable = True
        return result


def _resources_for_damage(resource: SimpleResourceInfo, database: ResourceDatabase) -> Iterator[ResourceInfo]:
    yield database.energy_tank
    for reduction in database.damage_reductions.get(resource, []):
        if reduction.inventory_item is not None:
            yield reduction.inventory_item


def calculate_interesting_resources(satisfiable_requirements: SatisfiableRequirements,
                                    resources: CurrentResources,
                                    energy: int,
                                    database: ResourceDatabase) -> FrozenSet[ResourceInfo]:
    """A resource is considered interesting if it isn't satisfied and it belongs to any satisfiable RequirementList """

    def helper():
        # For each possible requirement list
        for requirement_list in satisfiable_requirements:
            # If it's not satisfied, there's at least one IndividualRequirement in it that can be collected
            if not requirement_list.satisfied(resources, energy, database):

                for individual in requirement_list.values():
                    # Ignore those with the `negate` flag. We can't "uncollect" a resource to satisfy these.
                    # Finally, if it's not satisfied then we're interested in collecting it
                    if not individual.negate and not individual.satisfied(resources, energy, database):
                        if individual.is_damage:
                            yield from _resources_for_damage(individual.resource, database)
                        else:
                            yield individual.resource

    return frozenset(helper())



from construct import PrefixedArray, VarInt, Struct, CString, Flag

from randovania.games.binary_data import OptionalValue

BinStr = CString("utf-8")

BinaryInventory = PrefixedArray(
    VarInt,
    Struct(
        index=VarInt,
        amount=VarInt,
        capacity=VarInt,
    )
)

BinaryPlayerSessionEntry = Struct(
    id=VarInt,
    name=BinStr,
    row=OptionalValue(VarInt),
    admin=Flag,
    connection_state=BinStr,
)

BinaryGameDetails = Struct(
    seed_hash=BinStr,
    word_hash=BinStr,
    spoiler=Flag,
    permalink=BinStr,
)

BinaryGameSessionEntry = Struct(
    id=VarInt,
    name=BinStr,
    presets=PrefixedArray(VarInt, BinStr),
    players=PrefixedArray(VarInt, BinaryPlayerSessionEntry),
    game_details=OptionalValue(BinaryGameDetails),
    state=BinStr,
    generation_in_progress=OptionalValue(VarInt),
    allowed_games=PrefixedArray(VarInt, BinStr),
)

BinaryGameSessionAction = Struct(
    location=VarInt,
    pickup=BinStr,
    provider=BinStr,
    provider_row=VarInt,
    receiver=BinStr,
    time=BinStr,
)
BinaryGameSessionActions = PrefixedArray(VarInt, BinaryGameSessionAction)

BinaryGameSessionAudit = Struct(
    user=BinStr,
    message=BinStr,
    time=BinStr,
)
BinaryGameSessionAuditLog = PrefixedArray(VarInt, BinaryGameSessionAudit)



import collections
import math
from typing import Dict, List

from randovania.game_description.game_patches import GamePatches
from randovania.game_description.resources.pickup_entry import PickupEntry
from randovania.games.prime.patcher_file_lib import hint_lib
from randovania.interface_common.players_configuration import PlayersConfiguration
from randovania.layout.base.major_items_configuration import MajorItemsConfiguration


def locations_for_major_pickups_and_keys(
        all_patches: Dict[int, GamePatches],
        players_config: PlayersConfiguration,
        area_namers: Dict[int, hint_lib.AreaNamer],
) -> Dict[PickupEntry, List[str]]:
    """

    :param all_patches:
    :param players_config:
    :param area_namers:
    :return:
    """

    results: Dict[PickupEntry, List[str]] = collections.defaultdict(list)

    def make_det(index):
        return hint_lib.player_determiner(players_config, index).s

    for player_index, patches in all_patches.items():
        for pickup_index, target in patches.pickup_assignment.items():
            if target.player != players_config.player_index:
                continue

            item_category = target.pickup.item_category
            if item_category.is_major or item_category.is_key:
                hint = area_namers[player_index].location_name(pickup_index, hide_area=False, color=None)
                if players_config.is_multiworld:
                    hint = f"{make_det(player_index)}{hint}"

                results[target.pickup].append(hint)

    return results


def prime_trilogy_credits(
        major_items_configuration: MajorItemsConfiguration,
        all_patches: Dict[int, GamePatches],
        players_config: PlayersConfiguration,
        area_namers: Dict[int, hint_lib.AreaNamer],
        title: str,
        pickup_name_format: str,
) -> str:
    major_name_order = {
        pickup.name: index
        for index, pickup in enumerate(major_items_configuration.items_state.keys())
    }

    def sort_pickup(p: PickupEntry):
        return major_name_order.get(p.name, math.inf), p.name

    major_pickups_spoiler = locations_for_major_pickups_and_keys(all_patches,
                                                                 players_config,
                                                                 area_namers)

    credits_lines = [
        "{}\n{}".format(
            pickup_name_format.format(pickup.name),
            "\n".join(major_pickups_spoiler[pickup]) or "Nowhere"
        )
        for pickup in sorted(major_pickups_spoiler.keys(), key=sort_pickup)
    ]
    credits_lines.insert(0, title)
    return "\n\n".join(credits_lines)



import dataclasses
from typing import List

from randovania.bitpacking.bitpacking import BitPackDataclass
from randovania.bitpacking.json_dataclass import JsonDataclass
from randovania.bitpacking.type_enforcement import DataclassPostInitTypeCheck
from randovania.games import default_data
from randovania.games.game import RandovaniaGame
from randovania.layout.base.ammo_configuration import AmmoConfiguration
from randovania.layout.base.available_locations import AvailableLocationsConfiguration
from randovania.layout.base.damage_strictness import LayoutDamageStrictness
from randovania.layout.base.logical_resource_action import LayoutLogicalResourceAction
from randovania.layout.base.major_items_configuration import MajorItemsConfiguration
from randovania.layout.base.pickup_model import PickupModelStyle, PickupModelDataSource
from randovania.layout.base.trick_level_configuration import TrickLevelConfiguration
from randovania.layout.lib import location_list


class StartingLocationList(location_list.LocationList):
    @classmethod
    def areas_list(cls, game: RandovaniaGame):
        return location_list.area_locations_with_filter(game, lambda area: area.valid_starting_location)


@dataclasses.dataclass(frozen=True)
class BaseConfiguration(BitPackDataclass, JsonDataclass, DataclassPostInitTypeCheck):
    trick_level: TrickLevelConfiguration
    starting_location: StartingLocationList
    available_locations: AvailableLocationsConfiguration
    major_items_configuration: MajorItemsConfiguration
    ammo_configuration: AmmoConfiguration
    damage_strictness: LayoutDamageStrictness
    pickup_model_style: PickupModelStyle
    pickup_model_data_source: PickupModelDataSource
    multi_pickup_placement: bool
    logical_resource_action: LayoutLogicalResourceAction

    @classmethod
    def game_enum(cls) -> RandovaniaGame:
        raise NotImplementedError()

    @property
    def game(self):
        return self.game_enum()

    @property
    def game_data(self) -> dict:
        return default_data.read_json_then_binary(self.game)[1]

    @classmethod
    def json_extra_arguments(cls):
        return {
            "game": cls.game_enum(),
        }

    def dangerous_settings(self) -> List[str]:
        result = []
        for field in dataclasses.fields(self):
            f = getattr(self, field.name)
            if hasattr(f, "dangerous_settings"):
                result.extend(f.dangerous_settings())

        if self.multi_pickup_placement:
            result.append("Multi-pickup placement enabled")

        return result



import functools
import logging
import math
from pathlib import Path
from typing import List, Dict, Optional, NamedTuple

import discord
import graphviz
from discord import Embed
from discord.ext import commands
from discord_slash import SlashContext, SlashCommandOptionType, ComponentContext, ButtonStyle
from discord_slash.utils import manage_commands, manage_components

from randovania.game_description import default_database, pretty_print
from randovania.game_description.game_description import GameDescription
from randovania.game_description.world.area import Area
from randovania.game_description.world.world import World
from randovania.games.game import RandovaniaGame
from randovania.lib import enum_lib
from randovania.server.discord.bot import RandovaniaBot


class SplitWorld(NamedTuple):
    world: World
    name: str
    areas: List[Area]


def render_area_with_graphviz(area: Area) -> Optional[Path]:
    dot = graphviz.Digraph(comment=area.name)
    for node in area.nodes:
        dot.node(node.name)

    known_edges = set()
    for source, target in area.connections.items():
        for target_node, requirement in target.items():
            direction = None
            if source in area.connections.get(target_node):
                direction = "both"
                known_edges.add((target_node.name, source.name))

            if (source.name, target_node.name) not in known_edges:
                dot.edge(source.name, target_node.name, dir=direction)
                known_edges.add((source.name, target_node.name))

    try:
        return Path(dot.render(format="png", cleanup=True))
    except graphviz.backend.ExecutableNotFound as e:
        logging.info("Unable to render graph for %s: %s", area.name, str(e))
        return None


async def create_split_worlds(db: GameDescription) -> List[SplitWorld]:
    world_options = []

    for world in db.world_list.worlds:
        for is_dark_world in [False, True]:
            if is_dark_world and world.dark_name is None:
                continue

            areas = sorted([area for area in world.areas if area.in_dark_aether == is_dark_world],
                           key=lambda it: it.name)
            name = world.correct_name(is_dark_world)
            if len(areas) > 25:
                per_part = math.ceil(len(areas) / math.ceil(len(areas) / 25))

                while areas:
                    areas_part = areas[:per_part]
                    del areas[:per_part]

                    world_options.append(SplitWorld(world, "{} ({}-{})".format(name, areas_part[0].name[:2],
                                                                               areas_part[-1].name[:2]), areas_part))
            else:
                world_options.append(SplitWorld(world, name, areas))

    world_options.sort(key=lambda it: it[1])
    return world_options


class DatabaseCommandCog(commands.Cog):
    _split_worlds: Dict[RandovaniaGame, List[SplitWorld]]

    def __init__(self, configuration: dict, bot: RandovaniaBot):
        self.configuration = configuration
        self.bot = bot
        self._split_worlds = {}
        self._on_database_component_listener = {}

    @commands.Cog.listener()
    async def on_ready(self):
        self.bot.slash.add_slash_command(
            self.database_command,
            name="database-inspect",
            description="Consult the Randovania's logic database for one specific room.",
            guild_ids=[self.configuration["guild"]],
            options=[
                manage_commands.create_option(
                    "game", "The game's database to check.",
                    option_type=SlashCommandOptionType.STRING,
                    required=True,
                    choices=[manage_commands.create_choice(game.value, game.long_name)
                             for game in enum_lib.iterate_enum(RandovaniaGame)]
                )
            ],
        )

        def add_id(custom_id: str, call, **kwargs):
            self._on_database_component_listener[custom_id] = functools.partial(call, **kwargs)

        for game in enum_lib.iterate_enum(RandovaniaGame):
            db = default_database.game_description_for(game)
            world_options = await create_split_worlds(db)
            self._split_worlds[game] = world_options

            add_id(f"{game.value}_world", self.on_database_world_selected, game=game)
            add_id(f"back_to_{game.value}", self.on_database_back_to_game, game=game)

            for i, split_world in enumerate(world_options):
                add_id(f"{game.value}_world_{i}", self.on_database_area_selected, game=game, split_world=split_world,
                       world_id=i)
                for j, area in enumerate(split_world.areas):
                    add_id(f"{game.value}_world_{i}_area_{j}", self.on_area_node_selection, game=game, area=area)

        self.bot.slash.add_component_callback(
            self.on_database_component,
            components=list(self._on_database_component_listener.keys()),
            use_callback_name=False,
        )

        await self.bot.slash.sync_all_commands()

    def _message_components_for_game(self, game: RandovaniaGame):
        options = []
        for i, (world, world_name, areas) in enumerate(self._split_worlds[game]):
            custom_id = f"{game.value}_world_{i}"
            options.append(manage_components.create_select_option(world_name, value=custom_id))

        action_row = manage_components.create_actionrow(manage_components.create_select(
            options=options,
            custom_id=f"{game.value}_world",
            placeholder="Choose your region",
        ))
        embed = Embed(title=f"{game.long_name} Database", description="Choose the world subset to visualize.")
        logging.info("Responding requesting list of worlds for game %s.", game.long_name)
        return embed, [action_row],

    async def database_command(self, ctx: SlashContext, game: str):
        game = RandovaniaGame(game)
        embed, components = self._message_components_for_game(game)
        await ctx.send(embed=embed, components=components, hidden=True)

    async def on_database_component(self, ctx: ComponentContext):
        await self._on_database_component_listener[ctx.custom_id](ctx)

    async def on_database_back_to_game(self, ctx: ComponentContext, game: RandovaniaGame):
        embed, components = self._message_components_for_game(game)
        await ctx.edit_origin(embed=embed, components=components, hidden=True)

    async def on_database_world_selected(self, ctx: ComponentContext, game: RandovaniaGame):
        option_selected = ctx.selected_options[0]

        valid_items = [
            it
            for i, it in enumerate(self._split_worlds[game])
            if f"{game.value}_world_{i}" == option_selected
        ]
        if not valid_items:
            return await ctx.edit_origin(
                components=[], embeds=[],
                content=f"Invalid selected option, unable to find given world subset '{option_selected}'."
            )
        index = self._split_worlds[game].index(valid_items[0])
        world, world_name, areas = valid_items[0]

        select = manage_components.create_select(
            options=[
                manage_components.create_select_option(area.name, value=f"area_{i}")
                for i, area in sorted(enumerate(areas), key=lambda it: it[1].name)
            ],
            custom_id=f"{game.value}_world_{index}",
            placeholder="Choose the room",
        )
        action_row = manage_components.create_actionrow(select)
        back_row = manage_components.create_actionrow(manage_components.create_button(
            style=ButtonStyle.secondary,
            label="Back",
            custom_id=f"back_to_{game.value}",
        ))

        embed = Embed(title=f"{game.long_name} Database",
                      description=f"Choose the room in {world_name} to visualize.")
        logging.info("Responding to area selection for section %s with %d options.", world_name, len(areas))
        await ctx.edit_origin(
            embed=embed,
            components=[
                action_row,
                back_row,
            ],
        )

    async def on_database_area_selected(self, ctx: ComponentContext, game: RandovaniaGame, split_world: SplitWorld,
                                        world_id: int):
        option_selected = ctx.selected_options[0]

        valid_items = [
            area
            for i, area in enumerate(split_world.areas)
            if f"area_{i}" == option_selected
        ]
        if not valid_items:
            return await ctx.edit_origin(
                components=[], embeds=[],
                content=f"Invalid selected option, unable to find given area '{option_selected}'."
            )

        area = valid_items[0]
        db = default_database.game_description_for(game)

        embed = Embed(title="{}: {}".format(game.long_name, db.world_list.area_name(area)))

        select = manage_components.create_select(
            options=[
                manage_components.create_select_option(node.name, value=node.name)
                for i, node in sorted(enumerate(area.nodes), key=lambda it: it[1].name)
            ],
            max_values=len(area.nodes),
            custom_id=f"{game.value}_world_{world_id}_{option_selected}",
            placeholder="Choose the room",
        )
        action_row = manage_components.create_actionrow(select)

        files = []

        image_path = render_area_with_graphviz(area)
        if image_path is not None:
            files.append(discord.File(image_path))

        logging.info("Responding to area for %s with %d attachments.", area.name, len(files))
        await ctx.send(
            content=f"Requested by {ctx.author.display_name}.",
            embed=embed,
            files=files,
            components=[
                action_row,
            ],
        )

        if image_path is not None:
            image_path.unlink()

    async def on_area_node_selection(self, ctx: ComponentContext, game: RandovaniaGame, area: Area):
        db = default_database.game_description_for(game)

        embed = discord.Embed(title=ctx.origin_message.embeds[0].title)

        for i, node in enumerate(area.nodes):
            if node.name not in ctx.selected_options:
                continue

            name = node.name
            if node.heal:
                name += " (Heals)"
            if area.default_node_index == i:
                name += "; Spawn Point"

            body = pretty_print.pretty_print_node_type(node, db.world_list) + "\n"

            for target_node, requirement in db.world_list.area_connections_from(node):
                extra_lines = []
                for level, text in pretty_print.pretty_print_requirement(requirement.simplify()):
                    extra_lines.append("{}{}".format("  " * level, text))

                inner = "\n".join(extra_lines)
                new_entry = f"\n{target_node.name}:\n```\n{inner}\n```"

                if len(body) + len(new_entry) < 1024:
                    body += new_entry
                else:
                    logging.warning("Unable to add new entry: %d", len(new_entry))

            embed.add_field(
                name=name,
                value=body,
                inline=False,
            )

        logging.info("Updating visible nodes of %s: %s", area.name, str(ctx.selected_options))
        await ctx.edit_origin(
            embed=embed,
        )


def setup(bot: RandovaniaBot):
    bot.add_cog(DatabaseCommandCog(bot.configuration, bot))



# -*- coding: utf-8 -*-

################################################################################
## Form generated from reading UI file 'trick_usage_popup.ui'
##
## Created by: Qt User Interface Compiler version 5.15.2
##
## WARNING! All changes made in this file will be lost when recompiling UI file!
################################################################################

from PySide2.QtCore import *
from PySide2.QtGui import *
from PySide2.QtWidgets import *


class Ui_TrickUsagePopup(object):
    def setupUi(self, TrickUsagePopup):
        if not TrickUsagePopup.objectName():
            TrickUsagePopup.setObjectName(u"TrickUsagePopup")
        TrickUsagePopup.resize(431, 444)
        sizePolicy = QSizePolicy(QSizePolicy.Preferred, QSizePolicy.Preferred)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(TrickUsagePopup.sizePolicy().hasHeightForWidth())
        TrickUsagePopup.setSizePolicy(sizePolicy)
        TrickUsagePopup.setMaximumSize(QSize(16777215, 16777215))
        self.verticalLayout = QVBoxLayout(TrickUsagePopup)
        self.verticalLayout.setSpacing(6)
        self.verticalLayout.setContentsMargins(11, 11, 11, 11)
        self.verticalLayout.setObjectName(u"verticalLayout")
        self.title_label = QLabel(TrickUsagePopup)
        self.title_label.setObjectName(u"title_label")
        self.title_label.setWordWrap(True)
        self.title_label.setOpenExternalLinks(True)

        self.verticalLayout.addWidget(self.title_label)

        self.scroll_area = QScrollArea(TrickUsagePopup)
        self.scroll_area.setObjectName(u"scroll_area")
        self.scroll_area.setWidgetResizable(True)
        self.scroll_area_contents = QWidget()
        self.scroll_area_contents.setObjectName(u"scroll_area_contents")
        self.scroll_area_contents.setGeometry(QRect(0, 0, 411, 295))
        self.verticalLayout_2 = QVBoxLayout(self.scroll_area_contents)
        self.verticalLayout_2.setSpacing(6)
        self.verticalLayout_2.setContentsMargins(11, 11, 11, 11)
        self.verticalLayout_2.setObjectName(u"verticalLayout_2")
        self.area_list_label = QLabel(self.scroll_area_contents)
        self.area_list_label.setObjectName(u"area_list_label")
        self.area_list_label.setAlignment(Qt.AlignLeading|Qt.AlignLeft|Qt.AlignTop)
        self.area_list_label.setWordWrap(True)

        self.verticalLayout_2.addWidget(self.area_list_label)

        self.scroll_area.setWidget(self.scroll_area_contents)

        self.verticalLayout.addWidget(self.scroll_area)

        self.button_box = QDialogButtonBox(TrickUsagePopup)
        self.button_box.setObjectName(u"button_box")
        self.button_box.setStandardButtons(QDialogButtonBox.Close)

        self.verticalLayout.addWidget(self.button_box)


        self.retranslateUi(TrickUsagePopup)

        QMetaObject.connectSlotsByName(TrickUsagePopup)
    # setupUi

    def retranslateUi(self, TrickUsagePopup):
        TrickUsagePopup.setWindowTitle(QCoreApplication.translate("TrickUsagePopup", u"Trick Usage", None))
        self.title_label.setText(QCoreApplication.translate("TrickUsagePopup", u"<html><head/><body><p>This popup lists all occurences of tricks that are in logical use for this preset.</p><p>Enabled tricks:<br/>{trick_levels}</p><p>Click a room name to open it in the Data Visualizer for more details.</p></body></html>", None))
        self.area_list_label.setText(QCoreApplication.translate("TrickUsagePopup", u"<html><head/><body><p><a href=\"data-editor://World/Area\">INSERT AREAS!</a></p></body></html>", None))
    # retranslateUi




import collections
import copy
import dataclasses
from typing import Dict, Iterator, Tuple, List

from randovania.bitpacking import bitpacking
from randovania.bitpacking.bitpacking import BitPackValue, BitPackDecoder
from randovania.game_description import default_database
from randovania.game_description.resources.resource_database import ResourceDatabase
from randovania.game_description.resources.trick_resource_info import TrickResourceInfo
from randovania.games.game import RandovaniaGame
from randovania.layout.base.trick_level import LayoutTrickLevel
from randovania.lib.enum_lib import iterate_enum


def _all_tricks(resource_database: ResourceDatabase):
    return resource_database.trick


@dataclasses.dataclass(frozen=True)
class TrickLevelConfiguration(BitPackValue):
    minimal_logic: bool
    specific_levels: Dict[str, LayoutTrickLevel]
    game: RandovaniaGame

    def __post_init__(self):
        for trick, level in self.specific_levels.items():
            if not isinstance(level, LayoutTrickLevel) or level == LayoutTrickLevel.DISABLED:
                raise ValueError(f"Invalid level `{level}` for trick {trick}, "
                                 f"expected a LayoutTrickLevel that isn't NO_TRICKS")

    def bit_pack_encode(self, metadata) -> Iterator[Tuple[int, int]]:
        resource_database = default_database.resource_database_for(self.game)

        yield from bitpacking.encode_bool(self.minimal_logic)
        if self.minimal_logic:
            return

        encodable_levels = list(LayoutTrickLevel)
        encodable_levels.remove(LayoutTrickLevel.DISABLED)

        for trick in sorted(_all_tricks(resource_database)):
            has_trick = self.has_specific_level_for_trick(trick)
            yield from bitpacking.encode_bool(has_trick)
            if has_trick:
                yield from bitpacking.pack_array_element(self.level_for_trick(trick), encodable_levels)

    @classmethod
    def bit_pack_unpack(cls, decoder: BitPackDecoder, metadata):
        game = metadata["reference"].game
        resource_database = default_database.resource_database_for(game)

        minimal_logic = bitpacking.decode_bool(decoder)
        specific_levels = {}

        if not minimal_logic:
            encodable_levels = list(LayoutTrickLevel)
            encodable_levels.remove(LayoutTrickLevel.DISABLED)

            for trick in sorted(_all_tricks(resource_database)):
                if bitpacking.decode_bool(decoder):
                    specific_levels[trick.short_name] = decoder.decode_element(encodable_levels)

        return cls(minimal_logic, specific_levels, game)

    @property
    def pretty_description(self) -> str:
        if self.minimal_logic:
            return "Minimal Logic"

        count_at_difficulties = collections.defaultdict(list)
        for trick in _all_tricks(default_database.resource_database_for(self.game)):
            count_at_difficulties[self.level_for_trick(trick)].append(trick.long_name)

        if len(count_at_difficulties) == 1:
            for level in count_at_difficulties.keys():
                if level == LayoutTrickLevel.DISABLED:
                    return "All tricks disabled"
                return f"All tricks enabled at {level.long_name}"

        def tricks_at_level(tricks: List[str]) -> str:
            if len(tricks) != 1:
                return f"{len(tricks)}"
            else:
                return tricks[0]

        descriptions = [
            f"{tricks_at_level(count_at_difficulties[level])} at {level.long_name}"
            for level in iterate_enum(LayoutTrickLevel)
            if count_at_difficulties[level]
        ]
        return "Enabled tricks: {}".format(", ".join(descriptions))

    @property
    def as_json(self) -> dict:
        specific_levels = {
            trick_short_name: level.value
            for trick_short_name, level in sorted(self.specific_levels.items())
        }

        return {
            "minimal_logic": self.minimal_logic,
            "specific_levels": {} if self.minimal_logic else specific_levels,
        }

    @classmethod
    def from_json(cls, value: dict, game: RandovaniaGame):
        minimal_logic = value["minimal_logic"]
        specific_levels = {
            trick_short_name: LayoutTrickLevel(level)
            for trick_short_name, level in value["specific_levels"].items()
            if level != LayoutTrickLevel.DISABLED.value
        }
        return cls(
            minimal_logic=minimal_logic,
            specific_levels={} if minimal_logic else specific_levels,
            game=game,
        )

    def has_specific_level_for_trick(self, trick: TrickResourceInfo) -> bool:
        return trick.short_name in self.specific_levels

    def level_for_trick(self, trick: TrickResourceInfo) -> LayoutTrickLevel:
        return self.specific_levels.get(trick.short_name, LayoutTrickLevel.DISABLED)

    def set_level_for_trick(self, trick: TrickResourceInfo, value: LayoutTrickLevel) -> "TrickLevelConfiguration":
        """
        Creates a new TrickLevelConfiguration with the given trick with a changed level
        :param trick:
        :param value:
        :return:
        """
        assert value is not None
        new_levels = copy.copy(self.specific_levels)

        if value != LayoutTrickLevel.DISABLED:
            new_levels[trick.short_name] = value
        elif trick.short_name in new_levels:
            del new_levels[trick.short_name]

        return dataclasses.replace(self, specific_levels=new_levels)

    def dangerous_settings(self) -> List[str]:
        if self.minimal_logic:
            return ["Minimal Logic"]
        return []



import pytest
from mock import AsyncMock

from randovania.game_connection.connector.corruption_remote_connector import CorruptionRemoteConnector
from randovania.game_connection.executor.memory_operation import MemoryOperation


@pytest.fixture(name="connector")
def corruption_remote_connector():
    from randovania.games.prime import corruption_dol_versions
    connector = CorruptionRemoteConnector(corruption_dol_versions.ALL_VERSIONS[0])
    return connector


@pytest.mark.parametrize("correct_vtable", [False, True])
@pytest.mark.parametrize("has_cplayer", [False, True])
@pytest.mark.parametrize("has_pending_op", [False, True])
@pytest.mark.parametrize("has_world", [False, True])
@pytest.mark.asyncio
async def test_fetch_game_status(connector: CorruptionRemoteConnector, has_world, has_pending_op,
                                 has_cplayer, correct_vtable):
    # Setup
    expected_world = connector.game.world_list.worlds[1]

    cplayer_address = 0x8099FFAA

    executor = AsyncMock()
    executor.perform_memory_operations.side_effect = lambda ops: {
        ops[0]: expected_world.world_asset_id.to_bytes(8, "big") if has_world else b"DEADBEEF",
        ops[1]: b"\x01" if has_pending_op else b"\x00",
        ops[2]: cplayer_address.to_bytes(4, "big") if has_cplayer else None,
    }

    if correct_vtable:
        vtable_memory_return = connector.version.cplayer_vtable.to_bytes(4, "big")
    else:
        vtable_memory_return = b"CAFE"
    executor.perform_single_memory_operation.return_value = vtable_memory_return

    # Run
    actual_has_op, actual_world = await connector.current_game_status(executor)

    # Assert
    if has_world and has_cplayer and correct_vtable:
        assert actual_world is expected_world
    else:
        assert actual_world is None
    assert actual_has_op == has_pending_op
    if has_cplayer:
        executor.perform_single_memory_operation.assert_awaited_once_with(MemoryOperation(
            cplayer_address, read_byte_count=4,
        ))
    else:
        executor.perform_single_memory_operation.assert_not_awaited()
